"""Exception hook to filter out compiler internal frames from tracebacks."""

import sys
from types import TracebackType


def should_filter_traceback(tb: TracebackType | None) -> bool:
    return tb is not None and (
        tb.tb_frame.f_globals.get("_filter_traceback_", False) or should_filter_traceback(tb.tb_next)
    )


def is_compiler_internal(tb: TracebackType):
    return tb.tb_frame.f_locals.get("_compiler_internal_", False) or tb.tb_frame.f_globals.get(
        "_compiler_internal_", False
    )


def is_traceback_root(tb: TracebackType | None) -> bool:
    return tb.tb_frame.f_locals.get("_traceback_root_", False) or tb.tb_frame.f_globals.get("_traceback_root_", False)


def filter_traceback(tb: TracebackType | None) -> TracebackType | None:
    if tb is None:
        return None
    if is_compiler_internal(tb):
        return filter_traceback(tb.tb_next)
    tb.tb_next = filter_traceback(tb.tb_next)
    return tb


def truncate_traceback(tb: TracebackType | None) -> TracebackType | None:
    if tb is None:
        return None
    if is_compiler_internal(tb):
        return truncate_traceback(tb.tb_next)
    if is_traceback_root(tb):
        return tb
    else:
        return truncate_traceback(tb.tb_next)


def excepthook(exc, value, tb):
    import traceback

    if should_filter_traceback(tb):
        tb = filter_traceback(tb)
    traceback.print_exception(exc, value, tb)


def print_simple_traceback(exc, value, tb):
    import traceback

    if should_filter_traceback(tb):
        tb = filter_traceback(tb)
        truncated = truncate_traceback(tb)
        tb = truncated if truncated is not None else tb
        cause = value.__cause__
        value.__cause__ = None
        traceback.print_exception(exc, value, tb)
        value.__cause__ = cause
    else:
        traceback.print_exception(exc, value, tb)


def install_excepthook():
    sys.excepthook = excepthook



from math import pi
from typing import Self

from pydori.lib.options import Options
from sonolus.script.array import Dim
from sonolus.script.easing import ease_out_quad
from sonolus.script.globals import level_data
from sonolus.script.interval import clamp, interp_clamped, lerp, remap
from sonolus.script.quad import Quad, Rect
from sonolus.script.record import Record
from sonolus.script.transform import Transform2d
from sonolus.script.vec import Vec2

# Coordinate System Overview:
# - Many functions operate on an 'internal' coordinate system.
# - The internal coordinate system has the judge line at y=0 and notes approach it from a positive y value.
# - Notes are squares and lanes are rectangles in the internal coordinate system, making layout calculations easier.
# - Before being rendered, coordinates are transformed to the 'screen' coordinate system; this transformation
#   gives notes the appearance of approaching from an angle with 3D perspective.
# - Some layout calculations are done partially in the 'screen' coordinate system, such as arrow layout.

# Range of lanes in the game.
START_LANE = -3
END_LANE = 3
LANE_COUNT = 7
LANE_COUNT_DIM = Dim[7]

# The y-coordinate of the judge line in screen coordinates.
JUDGE_LINE_SCREEN_Y = -0.5

# The y-coordinate of the vanishing point in screen coordinates.
VANISHING_POINT_SCREEN_Y = 1.35

# Length of a lane at 100% lane length.
BASE_LANE_LENGTH = 8

# Width of a lane at 100% lane width.
BASE_LANE_WIDTH = 0.35

# Length over which notes fade in at the start and fade out at the end.
NOTE_FADE_LENGTH = 0.5

# Width of the stage border at 100% lane width.
BASE_STAGE_BORDER_WIDTH = BASE_LANE_WIDTH * 0.25

# Multiplier for directional flick arrow size compared to note width.
# Arrows are less visible than note bodies in typical skins, so we make them larger so they're easier to see.
DIRECTIONAL_FLICK_ARROW_SCALE = 1.5

# Initial and final vertical offset of the flick arrow during its animation cycle, relative to note width.
FLICK_ARROW_INITIAL_OFFSET = -0.1
FLICK_ARROW_FINAL_OFFSET = 0.4

# Horizontal offset of the directional flick arrow from note center.
# Negated for left flicks.
DIRECTIONAL_FLICK_OFFSET = 0.4

# Multiplier for circular effect size compared to note width.
CIRCULAR_EFFECT_SCALE = 1.8

# Multiplier for note hitbox width compared to lane width.
NOTE_HITBOX_SCALE = 2.5

# Scale factor for the note speed (note travel time).
# At this note speed, a note will take 1 second to travel from lane start to judge line.
# Higher values mean that the same speed will be slower / have a longer travel time.
REFERENCE_SPEED = 6


@level_data
class Layout:
    judge_line_screen_y: float
    vanishing_point_screen_y: float

    transform: Transform2d

    note_width: float
    lane_width: float
    stage_border_width: float

    note_y_max: float
    note_y_min: float


def init_layout():
    Layout.judge_line_screen_y = JUDGE_LINE_SCREEN_Y
    Layout.vanishing_point_screen_y = VANISHING_POINT_SCREEN_Y

    Layout.transform = Transform2d.new().perspective_y(
        Layout.judge_line_screen_y,
        Vec2(0, Layout.vanishing_point_screen_y),
    )

    Layout.note_width = BASE_LANE_WIDTH * Options.note_size
    Layout.lane_width = BASE_LANE_WIDTH * Options.lane_width
    Layout.stage_border_width = BASE_STAGE_BORDER_WIDTH

    Layout.note_y_max = Options.lane_length * BASE_LANE_LENGTH
    Layout.note_y_min = -1


def transform_vec(vec: Vec2) -> Vec2:
    """Apply perspective transformation to a vec."""
    return Layout.transform.transform_vec(vec)


def transform_quad(quad: Quad | Rect) -> Quad:
    """Apply perspective transformation to a quad."""
    return Layout.transform.transform_quad(quad)


def lane_to_x(lane: float) -> float:
    """Return the center x position of a lane."""
    return lane * Layout.lane_width


def lane_to_transformed_vec(lane: float) -> Vec2:
    """Return the transformed center x position of a lane at y=0."""
    return transform_vec(Vec2(lane_to_x(lane), 0))


def layout_lane(lane: float) -> Quad:
    center_x = lane_to_x(lane)
    return transform_quad(
        Rect(
            l=center_x - Layout.lane_width / 2,
            r=center_x + Layout.lane_width / 2,
            b=Layout.note_y_min,
            t=Layout.note_y_max,
        )
    )


def layout_stage_left_border() -> Quad:
    right = -LANE_COUNT / 2 * Layout.lane_width
    return transform_quad(
        Rect(
            l=right - Layout.stage_border_width,
            r=right,
            b=Layout.note_y_min,
            t=Layout.note_y_max,
        )
    )


def layout_stage_right_border() -> Quad:
    left = LANE_COUNT / 2 * Layout.lane_width
    return transform_quad(
        Rect(
            l=left,
            r=left + Layout.stage_border_width,
            b=Layout.note_y_min,
            t=Layout.note_y_max,
        )
    )


def layout_judge_line() -> Quad:
    return transform_quad(
        Rect(
            l=-LANE_COUNT / 2 * Layout.lane_width,
            r=LANE_COUNT / 2 * Layout.lane_width,
            b=-Layout.note_width / 2,
            t=Layout.note_width / 2,
        )
    )


def layout_note_body(lane: float, y: float) -> Quad:
    return transform_quad(
        Rect.from_center(
            Vec2(lane_to_x(lane), y),
            dimensions=Vec2(Layout.note_width, Layout.note_width),
        )
    )


def layout_flick_arrow(lane: float, y: float, progress: float) -> Quad:
    """Return the layout quad for a flick arrow.

    Args:
        lane: The lane number
        y: The y-coordinate of the arrow.
        progress: The progress of the flick animation (0 to 1).
    """
    base_bl = transform_vec(Vec2(lane_to_x(lane) - Layout.note_width / 2, y))
    base_br = transform_vec(Vec2(lane_to_x(lane) + Layout.note_width / 2, y))
    up = (base_br - base_bl).rotate(pi / 2)
    offset = lerp(FLICK_ARROW_INITIAL_OFFSET, FLICK_ARROW_FINAL_OFFSET, progress) * up
    bl = base_bl + offset
    br = base_br + offset
    tl = bl + up
    tr = br + up
    return Quad(bl=bl, br=br, tl=tl, tr=tr)


def layout_directional_flick_arrow(lane: float, y: float, direction: int, number: int, progress: float) -> Quad:
    """Return the layout quad for a directional flick arrow.

    Args:
        lane: The lane number
        y: The y-coordinate of the arrow.
        direction: The direction of the flick (positive for right, negative for left).
        number: The index of the arrow in the sequence ranging from 0 to (direction - 1) inclusive.
        progress: The progress of the flick animation (0 to 1).
    """
    direction_sign = 1 if direction > 0 else -1
    return transform_quad(
        Rect.from_center(
            center=Vec2(
                lane_to_x(lane + direction_sign * (DIRECTIONAL_FLICK_OFFSET + number + progress)),
                y,
            ),
            dimensions=Vec2(
                Layout.note_width * DIRECTIONAL_FLICK_ARROW_SCALE,
                Layout.note_width * DIRECTIONAL_FLICK_ARROW_SCALE,
            ),
        )
        .as_quad()
        .rotate_centered(
            # Rotate a quarter turn clockwise for right flicks, counterclockwise for left flicks.
            direction_sign * -pi / 2
        )
    )


def layout_note_linear_particle(lane: float) -> Quad:
    center_x = lane_to_x(lane)
    bl = transform_vec(Vec2(center_x - Layout.note_width / 2, 0))
    br = transform_vec(Vec2(center_x + Layout.note_width / 2, 0))
    # We want the effect to be as tall as it is wide
    height = (br - bl).magnitude
    tl = bl + Vec2(0, height)
    tr = br + Vec2(0, height)
    return Quad(bl=bl, br=br, tl=tl, tr=tr)


def layout_note_circular_particle(lane: float) -> Quad:
    center = Vec2(lane_to_x(lane), 0)
    result = transform_quad(
        Rect.from_center(
            center,
            dimensions=CIRCULAR_EFFECT_SCALE * Vec2(Layout.note_width, Layout.note_width),
        )
    )
    # Due to how particles are rendered, a perspective "trapezoid" shape will look weird,
    # so we make it a parallelogram instead.
    # Specifically, particles only render affine transformations correctly.
    mean_width = (result.tr.x - result.tl.x + result.br.x - result.bl.x) / 2
    mean_top_x = (result.tl.x + result.tr.x) / 2
    mean_bottom_x = (result.bl.x + result.br.x) / 2
    return Quad(
        bl=Vec2(mean_bottom_x - mean_width / 2, result.bl.y),
        br=Vec2(mean_bottom_x + mean_width / 2, result.br.y),
        tl=Vec2(mean_top_x - mean_width / 2, result.tl.y),
        tr=Vec2(mean_top_x + mean_width / 2, result.tr.y),
    )


def layout_hold_connector(
    lane_a: float,
    lane_b: float,
    y_a: float,
    y_b: float,
) -> Quad:
    y_a_adj = clamp(y_a, 0, Layout.note_y_max)
    y_b_adj = clamp(y_b, 0, Layout.note_y_max)
    lane_a_adj = remap(y_a, y_b, lane_a, lane_b, y_a_adj)
    lane_b_adj = remap(y_a, y_b, lane_a, lane_b, y_b_adj)
    center_x_a = lane_to_x(lane_a_adj)
    center_x_b = lane_to_x(lane_b_adj)
    return transform_quad(
        Quad(
            bl=Vec2(center_x_a - Layout.note_width / 2, y_a_adj),
            br=Vec2(center_x_a + Layout.note_width / 2, y_a_adj),
            tl=Vec2(center_x_b - Layout.note_width / 2, y_b_adj),
            tr=Vec2(center_x_b + Layout.note_width / 2, y_b_adj),
        )
    )


def layout_sim_line(
    lane_a: float,
    lane_b: float,
    y: float,
):
    left_x = lane_to_x(lane_a)
    right_x = lane_to_x(lane_b)
    return transform_quad(
        Rect(
            l=left_x,
            r=right_x,
            b=y - Layout.note_width / 2,
            t=y + Layout.note_width / 2,
        )
    )


def preempt_time() -> float:
    """Return the time between a note's spawn and when it reaches the judge line."""
    return REFERENCE_SPEED / Options.note_speed


def get_note_y(scaled_time: float, target_scaled_time: float) -> float:
    """Return the y-coordinate of a note at the given scaled time based on its target scaled time."""
    return remap(
        target_scaled_time - preempt_time(),
        target_scaled_time,
        Layout.note_y_max,
        0,
        scaled_time,
    )


def note_y_to_alpha(y: float) -> float:
    return ease_out_quad(
        interp_clamped(
            (
                Layout.note_y_min,
                Layout.note_y_min + NOTE_FADE_LENGTH,
                Layout.note_y_max - NOTE_FADE_LENGTH,
                Layout.note_y_max,
            ),
            (
                0,
                1,
                1,
                0,
            ),
            y,
        )
    )


class Hitbox(Record):
    """A hitbox bounded by a left and right edge."""

    left: float
    right: float

    @classmethod
    def from_center(cls, center: float, width: float) -> Self:
        left = center - width / 2
        right = center + width / 2
        return cls(left, right)

    @classmethod
    def for_note(cls, lane: float, direction: float) -> Self:
        result = cls.from_center(lane_to_x(lane), Layout.lane_width * NOTE_HITBOX_SCALE)
        if direction > 0:
            result.right += Layout.lane_width * direction
        if direction < 0:
            result.left += Layout.lane_width * direction
        return result

    def layout(self) -> Quad:
        return transform_quad(
            Rect(
                l=self.left,
                r=self.right,
                b=Layout.note_y_min,
                t=Layout.note_y_max,
            )
        )



from typing import Any

from sonolus.script.internal.dict_impl import DictImpl
from sonolus.script.internal.impl import validate_value
from sonolus.script.internal.meta_fn import meta_fn
from sonolus.script.record import Record


class SetImpl[Keys, OrderedKeys, Values](Record):
    _dict: DictImpl[Keys, OrderedKeys, Values]

    def __len__(self) -> int:
        return len(self._dict)

    def __contains__(self, item):
        return item in self._dict

    @meta_fn
    def __eq__(self, other: Any):
        raise TypeError("Set equality comparison is not supported")

    __hash__ = None

    @meta_fn
    def __or__(self, other):
        if not isinstance(other, SetImpl):
            raise TypeError("Unsupported type for '|' operator")
        return SetImpl(self._dict | other._dict)

    @staticmethod
    def from_set(s):
        values = [validate_value(v) for v in s]
        d = DictImpl.from_dict(dict.fromkeys(values))
        return SetImpl(d)



from math import cos, sin

from hypothesis import assume, given
from hypothesis import strategies as st
from hypothesis.stateful import RuleBasedStateMachine, invariant, precondition, rule

from sonolus.script.array import Array
from sonolus.script.interval import remap
from sonolus.script.transform import InvertibleTransform2d, Transform2d, perspective_approach
from sonolus.script.vec import Vec2
from tests.script.conftest import is_close, run_and_validate

# Use smaller limits to avoid precision issues.
# In practice, it's rare to see large arguments anyway
floats = st.floats(min_value=-9, max_value=9, allow_nan=False, allow_infinity=False)
nonzero_floats = floats.filter(lambda x: abs(x) > 1e-2)


@st.composite
def vecs(draw):
    x = draw(floats)
    y = draw(floats)
    return Vec2(x, y)


@st.composite
def vecs_nonzero(draw):
    x = draw(nonzero_floats)
    y = draw(nonzero_floats)
    return Vec2(x, y)


@given(
    v=vecs(),
    t=vecs(),
)
def test_translate(v, t):
    def fn():
        transform = Transform2d.new().translate(v)
        return transform.transform_vec(t)

    assert run_and_validate(fn) == v + t


@given(
    v=vecs(),
    factor=vecs(),
)
def test_scale(v, factor):
    def fn():
        transform = Transform2d.new().scale(factor)
        return transform.transform_vec(v)

    assert run_and_validate(fn) == v * factor


@given(
    v=vecs(),
    factor=vecs(),
    pivot=vecs(),
)
def test_scale_around(v, factor, pivot):
    def fn():
        transform = Transform2d.new().scale_about(factor, pivot)
        return transform.transform_vec(v)

    result = run_and_validate(fn)
    assert is_close(result.x, pivot.x + (v.x - pivot.x) * factor.x, rel_tol=1e-6, abs_tol=1e-6)
    assert is_close(result.y, pivot.y + (v.y - pivot.y) * factor.y, rel_tol=1e-6, abs_tol=1e-6)


@given(
    v=vecs(),
    angle=floats,
)
def test_rotate(v, angle):
    def fn():
        transform = Transform2d.new().rotate(angle)
        return transform.transform_vec(v)

    assert run_and_validate(fn) == v.rotate(angle)


@given(
    v=vecs(),
    angle=floats,
    pivot=vecs(),
)
def test_rotate_around(v, angle, pivot):
    def fn():
        transform = Transform2d.new().rotate_about(angle, pivot)
        return transform.transform_vec(v)

    result = run_and_validate(fn)
    assert is_close(
        result.x, pivot.x + (v.x - pivot.x) * cos(angle) - (v.y - pivot.y) * sin(angle), rel_tol=1e-6, abs_tol=1e-6
    )
    assert is_close(
        result.y, pivot.y + (v.x - pivot.x) * sin(angle) + (v.y - pivot.y) * cos(angle), rel_tol=1e-6, abs_tol=1e-6
    )


@given(
    v=vecs(),
    m=floats,
)
def test_shear_x(v, m):
    def fn():
        transform = Transform2d.new().shear_x(m)
        return transform.transform_vec(v)

    assert run_and_validate(fn) == Vec2(v.x + v.y * m, v.y)


@given(
    v=vecs(),
    m=floats,
)
def test_shear_y(v, m):
    def fn():
        transform = Transform2d.new().shear_y(m)
        return transform.transform_vec(v)

    assert run_and_validate(fn) == Vec2(v.x, v.x * m + v.y)


@given(
    v_x=floats,
    foreground_y=floats,
    vanishing_point=vecs(),
)
def test_perspective_y_at_foreground(v_x, foreground_y, vanishing_point):
    v = Vec2(v_x, 0)
    assume(abs(foreground_y - vanishing_point.y) > 1e-2)

    def fn():
        transform = Transform2d.new().perspective_y(foreground_y, vanishing_point)
        return transform.transform_vec(v)

    result = run_and_validate(fn)
    assert is_close(result.x, v_x, rel_tol=1e-4, abs_tol=1e-4)
    assert is_close(result.y, foreground_y, rel_tol=1e-4, abs_tol=1e-4)


@given(
    v_x=floats,
    foreground_y=floats,
    vanishing_point=vecs(),
)
def test_perspective_y_at_infinity(v_x, foreground_y, vanishing_point):
    v = Vec2(v_x, (vanishing_point.y - foreground_y) * 1e6)  # Close enough to infinity for testing
    assume(abs(foreground_y - vanishing_point.y) > 1e-2)

    def fn():
        transform = Transform2d.new().perspective_y(foreground_y, vanishing_point)
        return transform.transform_vec(v)

    result = run_and_validate(fn)
    assert is_close(result.x, vanishing_point.x, rel_tol=1e-3, abs_tol=1e-3)
    assert is_close(result.y, vanishing_point.y, rel_tol=1e-3, abs_tol=1e-3)


@given(
    v_y=floats,
    foreground_x=floats,
    vanishing_point=vecs(),
)
def test_perspective_x_at_foreground(v_y, foreground_x, vanishing_point):
    v = Vec2(0, v_y)
    assume(abs(foreground_x - vanishing_point.x) > 1e-2)

    def fn():
        transform = Transform2d.new().perspective_x(foreground_x, vanishing_point)
        return transform.transform_vec(v)

    result = run_and_validate(fn)
    assert is_close(result.x, foreground_x, rel_tol=1e-4, abs_tol=1e-4)
    assert is_close(result.y, v_y, rel_tol=1e-4, abs_tol=1e-4)


@given(
    v_y=floats,
    foreground_x=floats,
    vanishing_point=vecs(),
)
def test_perspective_x_at_infinity(v_y, foreground_x, vanishing_point):
    v = Vec2((vanishing_point.x - foreground_x) * 1e6, v_y)
    assume(abs(foreground_x - vanishing_point.x) > 1e-2)

    def fn():
        transform = Transform2d.new().perspective_x(foreground_x, vanishing_point)
        return transform.transform_vec(v)

    result = run_and_validate(fn)
    assert is_close(result.x, vanishing_point.x, rel_tol=1e-3, abs_tol=1e-3)
    assert is_close(result.y, vanishing_point.y, rel_tol=1e-3, abs_tol=1e-3)


class TransformInverse(RuleBasedStateMachine):
    def __init__(self):
        super().__init__()
        self.transform = Transform2d.new()
        self.inverse = Transform2d.new()

        # Some operations can lead to compounding precision error, so we limit the count
        self.scale_count = 0
        self.shear_count = 0
        self.perspective_count = 0

    @rule(translation=vecs())
    def translate(self, translation):
        self.transform = self.transform.translate(translation)
        self.inverse = Transform2d.new().translate(-translation).compose(self.inverse)

    @precondition(lambda self: self.scale_count < 2)
    @rule(factor=vecs_nonzero())
    def scale(self, factor):
        self.transform = self.transform.scale(factor)
        self.inverse = Transform2d.new().scale(Vec2.one() / factor).compose(self.inverse)
        self.scale_count += 1

    @precondition(lambda self: self.scale_count < 2)
    @rule(factor=vecs_nonzero(), pivot=vecs())
    def scale_about(self, factor, pivot):
        self.transform = self.transform.scale_about(factor, pivot)
        self.inverse = Transform2d.new().scale_about(Vec2.one() / factor, pivot).compose(self.inverse)
        self.scale_count += 1

    @rule(angle=floats)
    def rotate(self, angle):
        self.transform = self.transform.rotate(angle)
        self.inverse = Transform2d.new().rotate(-angle).compose(self.inverse)

    @rule(angle=floats, pivot=vecs())
    def rotate_about(self, angle, pivot):
        self.transform = self.transform.rotate_about(angle, pivot)
        self.inverse = Transform2d.new().rotate_about(-angle, pivot).compose(self.inverse)

    @precondition(lambda self: self.shear_count < 2)
    @rule(m=floats)
    def shear_x(self, m):
        self.transform = self.transform.shear_x(m)
        self.inverse = Transform2d.new().shear_x(-m).compose(self.inverse)
        self.shear_count += 1

    @precondition(lambda self: self.shear_count < 2)
    @rule(m=floats)
    def shear_y(self, m):
        self.transform = self.transform.shear_y(m)
        self.inverse = Transform2d.new().shear_y(-m).compose(self.inverse)
        self.shear_count += 1

    @precondition(lambda self: self.perspective_count < 2)
    @rule(y=nonzero_floats)
    def perspective_vanish_y(self, y):
        self.transform = self.transform.simple_perspective_y(y)
        self.inverse = Transform2d.new().simple_perspective_y(-y).compose(self.inverse)
        self.perspective_count += 1

    @precondition(lambda self: self.perspective_count < 2)
    @rule(foreground_y=floats, vanishing_point=vecs())
    def perspective_y(self, foreground_y, vanishing_point):
        assume(abs(foreground_y - vanishing_point.y) > 1e-2)
        self.transform = self.transform.perspective_y(foreground_y, vanishing_point)
        self.inverse = Transform2d.new().inverse_perspective_y(foreground_y, vanishing_point).compose(self.inverse)
        self.perspective_count += 1

    @precondition(lambda self: self.perspective_count < 2)
    @rule(foreground_x=floats, vanishing_point=vecs())
    def perspective_x(self, foreground_x, vanishing_point):
        assume(abs(foreground_x - vanishing_point.x) > 1e-2)
        self.transform = self.transform.perspective_x(foreground_x, vanishing_point)
        self.inverse = Transform2d.new().inverse_perspective_x(foreground_x, vanishing_point).compose(self.inverse)
        self.perspective_count += 1

    @invariant()
    def inverse_cancels(self):
        combo = self.transform.compose(self.inverse).normalize()
        abs_tol = 1e-2
        assert is_close(combo.a00, 1, abs_tol=abs_tol)
        assert is_close(combo.a01, 0, abs_tol=abs_tol)
        assert is_close(combo.a02, 0, abs_tol=abs_tol)
        assert is_close(combo.a10, 0, abs_tol=abs_tol)
        assert is_close(combo.a11, 1, abs_tol=abs_tol)
        assert is_close(combo.a12, 0, abs_tol=abs_tol)
        assert is_close(combo.a20, 0, abs_tol=abs_tol)
        assert is_close(combo.a21, 0, abs_tol=abs_tol)
        assert is_close(combo.a22, 1, abs_tol=abs_tol)


TestTransformInverse = TransformInverse.TestCase


@given(
    v=vecs(),
    t=vecs(),
)
def test_invertible_translate(v, t):
    def fn():
        transform = InvertibleTransform2d.new().translate(v)
        return transform.transform_vec(t)

    assert run_and_validate(fn) == v + t


@given(
    v=vecs(),
    t=vecs(),
)
def test_invertible_translate_inverse(v, t):
    def fn():
        transform = InvertibleTransform2d.new().translate(v)
        forward_result = transform.transform_vec(t)
        return transform.inverse_transform_vec(forward_result)

    result = run_and_validate(fn)
    assert is_close(result.x, t.x, rel_tol=1e-6, abs_tol=1e-6)
    assert is_close(result.y, t.y, rel_tol=1e-6, abs_tol=1e-6)


@given(
    v=vecs(),
    factor=vecs_nonzero(),
)
def test_invertible_scale(v, factor):
    def fn():
        transform = InvertibleTransform2d.new().scale(factor)
        return transform.transform_vec(v)

    assert run_and_validate(fn) == v * factor


@given(
    v=vecs(),
    factor=vecs_nonzero(),
)
def test_invertible_scale_inverse(v, factor):
    def fn():
        transform = InvertibleTransform2d.new().scale(factor)
        forward_result = transform.transform_vec(v)
        return transform.inverse_transform_vec(forward_result)

    result = run_and_validate(fn)
    assert is_close(result.x, v.x, rel_tol=1e-6, abs_tol=1e-6)
    assert is_close(result.y, v.y, rel_tol=1e-6, abs_tol=1e-6)


@given(
    v=vecs(),
    angle=floats,
)
def test_invertible_rotate(v, angle):
    def fn():
        transform = InvertibleTransform2d.new().rotate(angle)
        return transform.transform_vec(v)

    assert run_and_validate(fn) == v.rotate(angle)


@given(
    v=vecs(),
    angle=floats,
)
def test_invertible_rotate_inverse(v, angle):
    def fn():
        transform = InvertibleTransform2d.new().rotate(angle)
        forward_result = transform.transform_vec(v)
        return transform.inverse_transform_vec(forward_result)

    result = run_and_validate(fn)
    assert is_close(result.x, v.x, rel_tol=1e-6, abs_tol=1e-6)
    assert is_close(result.y, v.y, rel_tol=1e-6, abs_tol=1e-6)


@given(
    v=vecs(),
    m=floats,
)
def test_invertible_shear_x(v, m):
    def fn():
        transform = InvertibleTransform2d.new().shear_x(m)
        return transform.transform_vec(v)

    assert run_and_validate(fn) == Vec2(v.x + v.y * m, v.y)


@given(
    v=vecs(),
    m=floats,
)
def test_invertible_shear_x_inverse(v, m):
    def fn():
        transform = InvertibleTransform2d.new().shear_x(m)
        forward_result = transform.transform_vec(v)
        return transform.inverse_transform_vec(forward_result)

    result = run_and_validate(fn)
    assert is_close(result.x, v.x, rel_tol=1e-6, abs_tol=1e-6)
    assert is_close(result.y, v.y, rel_tol=1e-6, abs_tol=1e-6)


@given(
    v=vecs(),
    m=floats,
)
def test_invertible_shear_y(v, m):
    def fn():
        transform = InvertibleTransform2d.new().shear_y(m)
        return transform.transform_vec(v)

    assert run_and_validate(fn) == Vec2(v.x, v.x * m + v.y)


@given(
    v=vecs(),
    m=floats,
)
def test_invertible_shear_y_inverse(v, m):
    def fn():
        transform = InvertibleTransform2d.new().shear_y(m)
        forward_result = transform.transform_vec(v)
        return transform.inverse_transform_vec(forward_result)

    result = run_and_validate(fn)
    assert is_close(result.x, v.x, rel_tol=1e-6, abs_tol=1e-6)
    assert is_close(result.y, v.y, rel_tol=1e-6, abs_tol=1e-6)


@given(
    v=vecs(),
    translation=vecs(),
    factor=vecs_nonzero(),
    angle=floats,
)
def test_invertible_multiple_inverse(v, translation, factor, angle):
    def fn():
        transform = InvertibleTransform2d.new().translate(translation).scale(factor).rotate(angle)
        forward_result = transform.transform_vec(v)
        return transform.inverse_transform_vec(forward_result)

    result = run_and_validate(fn)
    assert is_close(result.x, v.x, rel_tol=1e-4, abs_tol=1e-4)
    assert is_close(result.y, v.y, rel_tol=1e-4, abs_tol=1e-4)


@given(
    v=vecs(),
    translation=vecs(),
    angle=floats,
)
def test_invertible_compose_matches_direct(v, translation, angle):
    # noinspection PyShadowingNames
    def fn():
        transform1 = InvertibleTransform2d.new().translate(translation)
        transform2 = InvertibleTransform2d.new().rotate(angle)
        transform_composed = transform1.compose(transform2)
        transform_direct = InvertibleTransform2d.new().translate(translation).rotate(angle)
        forward_result = transform_composed.transform_vec(v)
        forward_result_direct = transform_direct.transform_vec(v)
        inverse_result = transform_composed.inverse_transform_vec(forward_result)
        inverse_result_direct = transform_direct.inverse_transform_vec(forward_result_direct)
        return Array(
            forward_result,
            forward_result_direct,
            inverse_result,
            inverse_result_direct,
        )

    result = run_and_validate(fn)
    forward_result, forward_result_direct, inverse_result, inverse_result_direct = result
    assert is_close(forward_result.x, forward_result_direct.x, rel_tol=1e-4, abs_tol=1e-4)
    assert is_close(forward_result.y, forward_result_direct.y, rel_tol=1e-4, abs_tol=1e-4)
    assert is_close(inverse_result.x, inverse_result_direct.x, rel_tol=1e-4, abs_tol=1e-4)
    assert is_close(inverse_result.y, inverse_result_direct.y, rel_tol=1e-4, abs_tol=1e-4)


class InvertibleTransformStateMachine(RuleBasedStateMachine):
    def __init__(self):
        super().__init__()
        self.transform = InvertibleTransform2d.new()
        self.scale_count = 0
        self.shear_count = 0
        self.perspective_count = 0

    @rule(translation=vecs())
    def translate(self, translation):
        self.transform = self.transform.translate(translation)

    @precondition(lambda self: self.scale_count < 2)
    @rule(factor=vecs_nonzero())
    def scale(self, factor):
        self.transform = self.transform.scale(factor)
        self.scale_count += 1

    @precondition(lambda self: self.scale_count < 2)
    @rule(factor=vecs_nonzero(), pivot=vecs())
    def scale_about(self, factor, pivot):
        self.transform = self.transform.scale_about(factor, pivot)
        self.scale_count += 1

    @rule(angle=floats)
    def rotate(self, angle):
        self.transform = self.transform.rotate(angle)

    @rule(angle=floats, pivot=vecs())
    def rotate_about(self, angle, pivot):
        self.transform = self.transform.rotate_about(angle, pivot)

    @precondition(lambda self: self.shear_count < 2)
    @rule(m=floats)
    def shear_x(self, m):
        self.transform = self.transform.shear_x(m)
        self.shear_count += 1

    @precondition(lambda self: self.shear_count < 2)
    @rule(m=floats)
    def shear_y(self, m):
        self.transform = self.transform.shear_y(m)
        self.shear_count += 1

    @precondition(lambda self: self.perspective_count < 2)
    @rule(y=nonzero_floats)
    def simple_perspective_y(self, y):
        self.transform = self.transform.simple_perspective_y(y)
        self.perspective_count += 1

    @precondition(lambda self: self.perspective_count < 2)
    @rule(foreground_y=floats, vanishing_point=vecs())
    def perspective_y(self, foreground_y, vanishing_point):
        assume(abs(foreground_y - vanishing_point.y) > 1e-2)
        self.transform = self.transform.perspective_y(foreground_y, vanishing_point)
        self.perspective_count += 1

    @precondition(lambda self: self.perspective_count < 2)
    @rule(foreground_x=floats, vanishing_point=vecs())
    def perspective_x(self, foreground_x, vanishing_point):
        assume(abs(foreground_x - vanishing_point.x) > 1e-2)
        self.transform = self.transform.perspective_x(foreground_x, vanishing_point)
        self.perspective_count += 1

    @invariant()
    def inverse_cancels(self):
        combo = self.transform.forward.compose(self.transform.inverse).normalize()
        abs_tol = 1e-2
        assert is_close(combo.a00, 1, abs_tol=abs_tol)
        assert is_close(combo.a01, 0, abs_tol=abs_tol)
        assert is_close(combo.a02, 0, abs_tol=abs_tol)
        assert is_close(combo.a10, 0, abs_tol=abs_tol)
        assert is_close(combo.a11, 1, abs_tol=abs_tol)
        assert is_close(combo.a12, 0, abs_tol=abs_tol)
        assert is_close(combo.a20, 0, abs_tol=abs_tol)
        assert is_close(combo.a21, 0, abs_tol=abs_tol)
        assert is_close(combo.a22, 1, abs_tol=abs_tol)


TestInvertibleTransform = InvertibleTransformStateMachine.TestCase


def test_perspective_approach_at_0():
    def fn():
        return perspective_approach(2, 0)

    assert run_and_validate(fn) == 0


def test_perspective_approach_at_1():
    def fn():
        return perspective_approach(2, 1)

    assert run_and_validate(fn) == 1


def test_perspective_approach_at_halfway():
    def fn():
        return perspective_approach(2, 0.5)

    assert run_and_validate(fn) < 0.5


@given(
    progress=st.floats(min_value=-0.1, max_value=0.1),
)
def test_perspective_approach_against_transform(progress):
    def fn():
        transform = InvertibleTransform2d.new().perspective_y(0, Vec2(0, 2))
        y_0 = 10
        y_1 = 0
        w_0 = transform.transform_vec(Vec2(1, y_0)).x - transform.transform_vec(Vec2(0, y_0)).x
        w_1 = transform.transform_vec(Vec2(1, y_1)).x - transform.transform_vec(Vec2(0, y_1)).x
        d_0 = 1 / w_0
        d_1 = 1 / w_1
        y = remap(0, 1, y_0, y_1, progress)
        ty_0 = transform.transform_vec(Vec2(0, y_0)).y
        ty_1 = transform.transform_vec(Vec2(0, y_1)).y
        return Array(
            perspective_approach(d_0 / d_1, progress),
            remap(ty_0, ty_1, 0, 1, transform.transform_vec(Vec2(0, y)).y),
        )

    result = run_and_validate(fn)
    assert is_close(result[0], result[1], rel_tol=1e-4, abs_tol=1e-4)



from __future__ import annotations

from collections.abc import Iterator
from typing import Any, Final

from sonolus.script.internal.context import ctx
from sonolus.script.internal.meta_fn import meta_fn
from sonolus.script.maybe import Maybe, Nothing, Some
from sonolus.script.record import Record


class SonolusIterator[T]:
    """Base class for Sonolus iterators.

    This class is used to define custom iterators that can be used in Sonolus.py.

    Inheritors must implement the [`next`][sonolus.script.iterator.SonolusIterator.next] method,
    which should return a [`Maybe[T]`][sonolus.script.maybe.Maybe].

    Usage:
        ```python
        class MyIterator(Record, SonolusIterator):
            def next(self) -> Maybe[T]:
                ...
        ```
    """

    _allow_instance_check_ = True

    @meta_fn
    def next(self) -> Maybe[T]:
        """Return the next item from the iterator as a [`Maybe`][sonolus.script.maybe.Maybe]."""
        raise NotImplementedError

    def __next__(self) -> T:
        result = self.next()
        if result.is_some:
            return result.get_unsafe()
        else:
            raise StopIteration

    def __iter__(self) -> SonolusIterator[T]:
        return self


class _Enumerator[V: SonolusIterator](Record, SonolusIterator):
    i: int
    offset: Final[int]
    iterator: V

    def next(self) -> Maybe[tuple[int, Any]]:
        value = self.iterator.next()
        if value.is_nothing:
            return Nothing
        result = (self.i + self.offset, value.get_unsafe())
        self.i += 1
        return Some(result)


class _Zipper[T](Record, SonolusIterator):
    # Can be a, Pair[a, b], Pair[a, Pair[b, c]], etc.
    iterators: T

    @meta_fn
    def _get_iterators(self) -> tuple[SonolusIterator, ...]:
        from sonolus.script.containers import Pair

        iterators = []
        v = self.iterators
        while isinstance(v, Pair):
            iterators.append(v.first)
            v = v.second
        iterators.append(v)
        return tuple(iterators)

    @meta_fn
    def _get_next_values(self) -> tuple[Any, ...]:
        from sonolus.backend.visitor import compile_and_call

        return tuple(compile_and_call(iterator.next) for iterator in self._get_iterators())

    @meta_fn
    def _values_to_tuple(self, values: tuple[Any, ...]) -> tuple[Any, ...]:
        from sonolus.backend.visitor import compile_and_call

        return tuple(compile_and_call(value.get_unsafe) for value in values)

    def next(self) -> Maybe[tuple[Any, ...]]:
        values = self._get_next_values()
        for value in values:
            if value.is_nothing:
                return Nothing
        return Some(self._values_to_tuple(values))


class _EmptyIterator(Record, SonolusIterator):
    def next(self) -> Maybe[Any]:
        return Nothing


class _MappingIterator[T, Fn](Record, SonolusIterator):
    fn: Fn
    iterator: T

    def next(self) -> Maybe[Any]:
        return self.iterator.next().map(self.fn)


class _FilteringIterator[T, Fn](Record, SonolusIterator):
    fn: Fn
    iterator: T

    def next(self) -> Maybe[T]:
        while True:
            value = self.iterator.next()
            if value.is_nothing:
                return Nothing
            inside = value.get_unsafe()
            if self.fn(inside):
                return Some(inside)


@meta_fn
def maybe_next[T](iterator: Iterator[T]) -> Maybe[T]:
    """Get the next item from an iterator as a [`Maybe`][sonolus.script.maybe.Maybe]."""
    from sonolus.backend.visitor import compile_and_call

    if not isinstance(iterator, SonolusIterator):
        raise TypeError("Iterator must be an instance of SonolusIterator.")
    if ctx():
        return compile_and_call(iterator.next)
    else:
        return iterator.next()



# ruff: noqa: PLW0108
"""Tests for assert statements.

PYTEST_DONT_REWRITE
"""

import random

from sonolus.script.debug import assert_false, assert_true, notify, require
from sonolus.script.internal.context import RuntimeChecks
from tests.script.conftest import run_compiled


def test_assertion_succeeds():
    log_calls = []

    def fn():
        assert True, "Message"
        return 1

    assert run_compiled(fn, log_callback=lambda x: log_calls.append(x)) == 1
    assert len(log_calls) == 0


def test_assertion_fails():
    log_calls = []

    def fn():
        assert False, "Message"  # noqa: B011, PT015
        # noinspection PyUnreachableCode
        return 1

    result = run_compiled(
        fn, runtime_checks=RuntimeChecks.NOTIFY_AND_TERMINATE, log_callback=lambda x: log_calls.append(x)
    )
    assert result == 0
    assert len(log_calls) == 1


def test_assert_true_succeeds():
    log_calls = []

    def fn():
        assert_true(True)
        return 1

    assert run_compiled(fn, log_callback=lambda x: log_calls.append(x)) == 1
    assert len(log_calls) == 0


def test_assert_true_fails():
    log_calls = []

    def fn():
        assert_true(False)
        return 1

    result = run_compiled(
        fn, runtime_checks=RuntimeChecks.NOTIFY_AND_TERMINATE, log_callback=lambda x: log_calls.append(x)
    )
    assert result == 0
    assert len(log_calls) == 1


def test_assert_false_succeeds():
    log_calls = []

    def fn():
        assert_false(False)
        return 1

    assert run_compiled(fn, log_callback=lambda x: log_calls.append(x)) == 1
    assert len(log_calls) == 0


def test_assert_false_fails():
    log_calls = []

    def fn():
        assert_false(True)
        return 1

    result = run_compiled(
        fn, runtime_checks=RuntimeChecks.NOTIFY_AND_TERMINATE, log_callback=lambda x: log_calls.append(x)
    )
    assert result == 0
    assert len(log_calls) == 1


def test_assertion_fails_terminate_mode():
    log_calls = []

    def fn():
        assert False, "Message"  # noqa: B011, PT015
        # noinspection PyUnreachableCode
        return 1

    result = run_compiled(fn, runtime_checks=RuntimeChecks.TERMINATE, log_callback=lambda x: log_calls.append(x))
    assert result == 0
    assert len(log_calls) == 0


def test_assert_true_fails_terminate_mode():
    log_calls = []

    def fn():
        assert_true(False)
        return 1

    result = run_compiled(fn, runtime_checks=RuntimeChecks.TERMINATE, log_callback=lambda x: log_calls.append(x))
    assert result == 0
    assert len(log_calls) == 0


def test_assert_false_fails_terminate_mode():
    log_calls = []

    def fn():
        assert_false(True)
        return 1

    result = run_compiled(fn, runtime_checks=RuntimeChecks.TERMINATE, log_callback=lambda x: log_calls.append(x))
    assert result == 0
    assert len(log_calls) == 0


def test_assertion_dynamic_fails_none_mode():
    log_calls = []

    def fn():
        assert random.random() == -1, "Message"
        # noinspection PyUnreachableCode
        return 1

    assert run_compiled(fn, runtime_checks=RuntimeChecks.NONE, log_callback=lambda x: log_calls.append(x)) == 1
    assert len(log_calls) == 0


def test_assertion_static_fails_none_mode():
    log_calls = []

    def fn():
        assert False, "Message"  # noqa: B011, PT015
        # noinspection PyUnreachableCode
        return 1

    # Even in none, an assertion with a statically known-false condition should terminate.
    result = run_compiled(fn, runtime_checks=RuntimeChecks.TERMINATE, log_callback=lambda x: log_calls.append(x))
    assert result == 0
    assert len(log_calls) == 0


def test_notify_notify_and_terminate_mode():
    log_calls = []

    def fn():
        notify("Test notification")
        return 1

    assert (
        run_compiled(fn, runtime_checks=RuntimeChecks.NOTIFY_AND_TERMINATE, log_callback=lambda x: log_calls.append(x))
        == 1
    )
    assert len(log_calls) == 1


def test_notify_terminate_mode():
    log_calls = []

    def fn():
        notify("Test notification")
        return 1

    assert run_compiled(fn, runtime_checks=RuntimeChecks.TERMINATE, log_callback=lambda x: log_calls.append(x)) == 1
    assert len(log_calls) == 0


def test_notify_none_mode():
    log_calls = []

    def fn():
        notify("Test notification")
        return 1

    assert run_compiled(fn, runtime_checks=RuntimeChecks.NONE, log_callback=lambda x: log_calls.append(x)) == 1
    assert len(log_calls) == 0


def test_require_succeeds():
    log_calls = []

    def fn():
        require(True, "Should not fail")
        return 1

    assert run_compiled(fn, log_callback=lambda x: log_calls.append(x)) == 1
    assert len(log_calls) == 0


def test_require_fails():
    log_calls = []

    def fn():
        require(False, "Requirement failed")
        return 1

    result = run_compiled(
        fn, runtime_checks=RuntimeChecks.NOTIFY_AND_TERMINATE, log_callback=lambda x: log_calls.append(x)
    )
    assert result == 0
    assert len(log_calls) == 1


def test_require_fails_terminate_mode():
    log_calls = []

    def fn():
        require(False, "Requirement failed")
        return 1

    result = run_compiled(fn, runtime_checks=RuntimeChecks.TERMINATE, log_callback=lambda x: log_calls.append(x))
    assert result == 0
    assert len(log_calls) == 0


def test_require_fails_none_mode():
    log_calls = []

    def fn():
        require(False, "Requirement failed")
        return 1

    result = run_compiled(fn, runtime_checks=RuntimeChecks.NONE, log_callback=lambda x: log_calls.append(x))
    assert result == 0
    assert len(log_calls) == 0



from abc import abstractmethod
from collections.abc import Callable, Iterable
from types import NotImplementedType
from typing import Any, Self

from sonolus.backend.ir import IRConst, IRExpr, IRStmt
from sonolus.backend.place import BlockPlace


class BackingValue:
    def read(self) -> IRExpr:
        raise NotImplementedError()

    def write(self, value: IRExpr) -> IRStmt:
        raise NotImplementedError()


class ExprBackingValue(BackingValue):
    """A backing value that is backed by an expression."""

    def __init__(self, expr: IRExpr):
        self._expr = expr

    def read(self) -> IRExpr:
        return self._expr

    def write(self, value: IRExpr) -> IRStmt:
        raise RuntimeError("Value is read-only, cannot write to it")


class ReadOnlyBackingValueWrapper(BackingValue):
    """A wrapper around another backing value that blocks reads."""

    def __init__(self, value: BackingValue):
        self._value = value

    def read(self) -> IRExpr:
        return self._value.read()

    def write(self, value: IRExpr) -> IRStmt:
        raise RuntimeError("Value is read-only, cannot write to it")


type DataValue = BlockPlace | BackingValue | float | int | bool
type BackingSource = Callable[[IRExpr], BackingValue]


class Value:
    """Base class for values."""

    @classmethod
    def _is_concrete_(cls) -> bool:
        """Returns whether this type is concrete (i.e. can be instantiated)."""
        return False

    @classmethod
    @abstractmethod
    def _size_(cls) -> int:
        """Returns the size of this value."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _is_value_type_(cls) -> bool:
        """Returns whether this is a value type.

        If this is true, the value behaves immutably and _set_ is supported.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _from_place_(cls, place: BlockPlace) -> Self:
        """Creates a value from a place."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _accepts_(cls, value: Any) -> bool:
        """Returns whether this value can accept the given value."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _accept_(cls, value: Any) -> Self:
        """Accepts the given value."""
        raise NotImplementedError

    @abstractmethod
    def _is_py_(self) -> bool:
        """Returns whether this value is a valid Python value."""
        raise NotImplementedError

    @abstractmethod
    def _as_py_(self) -> Any:
        """Returns the Python equivalent of this value.

        Will fail if _is_py_ returns false.
        """
        raise NotImplementedError

    @classmethod
    def _from_backing_source_(cls, source: BackingSource) -> Self:
        """Creates a value from a backing source."""
        return cls._from_list_(source(IRConst(i)) for i in range(cls._size_()))

    @classmethod
    @abstractmethod
    def _from_list_(cls, values: Iterable[DataValue]) -> Self:
        """Creates a value from a list of data values."""
        raise NotImplementedError

    @abstractmethod
    def _to_list_(self, level_refs: dict[Any, str] | None = None) -> list[DataValue | str]:
        """Converts this value to a list of data values."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _flat_keys_(cls, prefix: str) -> list[str]:
        """Returns the keys to a flat representation of this value."""
        raise NotImplementedError

    def _to_flat_dict_(self, prefix: str, level_refs: dict[Any, str] | None = None) -> dict[str, DataValue | str]:
        """Converts this value to a flat dictionary."""
        return dict(zip(self._flat_keys_(prefix), self._to_list_(level_refs), strict=False))

    @abstractmethod
    def _get_(self) -> Self:
        """Implements access to the value.

        This is used when accessing a value from a record or array (and when storing into a record).
        For value (immutable) types, this makes a copy to preserve the appearance of immutability.
        For mutable types, this returns self.

        For instance:
        ```
        class X(Record):
            v: int

        a = 1
        b = X(a) # (1) _get_() is called on a
        c = b.v # (2) _get_() is called on the value for v

        # (1) prevents this from changing the value of a
        # (2) prevents this from changing the value of c
        # Thus, both calls to _get_() are necessary to ensure values behave immutably.
        b.v = 2
        ```
        """
        raise NotImplementedError

    def _get_readonly_(self) -> Self:
        """Implements access to the value without copying if the underlying value is immutable.

        The returned value should not be intentionally modified, but it is not guaranteed to be immutable.

        For example, a Num might be backed internally by rom, which is immutable. If we aren't going to modify
        (e.g. by putting it into a record where it can be modified), we can just return the original value, which
        avoids unnecessary copying.
        """
        return self._get_()

    @abstractmethod
    def _set_(self, value: Any):
        """Implements assignment (=).

        This is only supported by value types.
        This method must not change the active context like by branching.

        In some places, = might instead call assign_, such as when setting the value of a record field
        with a reference type.
        """
        raise NotImplementedError

    @abstractmethod
    def _copy_from_(self, value: Any):
        """Implements copy assignment (@=).

        This is only supported by mutable reference types.
        """
        raise NotImplementedError

    @abstractmethod
    def _copy_(self) -> Self:
        """Returns a deep copy of this value."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _alloc_(cls) -> Self:
        """Allocates a new value which may be uninitialized."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _zero_(cls) -> Self:
        """Returns a zero-initialized value of this type."""
        raise NotImplementedError

    @classmethod
    def _get_merge_target_(cls, values: list[Any]) -> Any | NotImplementedType:
        """Return the target when merging values from multiple code paths.

        E.g. for code like this:
        ```
        if cond:
            x = 1
        else:
            x = 2
        do_something(x)
        ```
        This is called to create a target value for x after the if-else block,
        and at the end of each block, that target value is assigned to the respective value (1 or 2).
        This lets us keep the value of x as a constant within if and else branches, and only have it
        become a runtime value after the if-else block.

        This is an overrideable method to allow for some other special behavior, namely the Maybe type.
        """
        if cls._is_value_type_():
            from sonolus.script.internal.context import ctx

            return cls._from_place_(ctx().alloc(size=cls._size_()))
        else:
            return NotImplemented

    def __imatmul__(self, other):
        self._copy_from_(other)
        return self


Value.__imatmul__._meta_fn_ = True  # type: ignore



from sonolus.script.array import Array
from sonolus.script.debug import debug_log
from tests.script.conftest import run_and_validate
from tests.script.test_record import Pair


def call_function(f, *args, **kwargs):
    return f(*args, **kwargs)


def test_simple_function_call():
    def a():
        return 1

    def fn():
        return a()

    assert run_and_validate(fn) == 1


def test_lambda_function_call():
    a = lambda: 1  # noqa: E731
    fn = lambda: a()  # noqa: PLW0108, E731

    assert run_and_validate(fn) == 1


def test_call_nested_function():
    def fn():
        def a():
            return 1

        return a()

    assert run_and_validate(fn) == 1


def test_indirect_call_nested_function():
    def fn():
        def a():
            return 1

        return call_function(a)

    assert run_and_validate(fn) == 1


def test_call_closure():
    def fn():
        x = 1

        def a():
            return x

        debug_log(a())
        debug_log(call_function(a))

        x = 2

        debug_log(a())
        debug_log(call_function(a))

        return a()

    assert run_and_validate(fn) == 2


def test_call_closure_with_default_args():
    def fn():
        x = 1

        def a(y=10):
            return x + y

        result = a()
        debug_log(result)
        return result

    assert run_and_validate(fn) == 11


def test_nested_lambda_closure():
    def fn():
        x = 1
        f = lambda y: x + y  # noqa: E731
        return f(2)

    assert run_and_validate(fn) == 3


def test_multiple_nested_functions():
    def fn():
        def outer(x):
            def middle(y):
                def inner(z):
                    return x + y + z

                return inner(3)

            return middle(2)

        return outer(1)

    assert run_and_validate(fn) == 6


def test_higher_order_function():
    def fn():
        def make_adder(x):
            return lambda y: x + y

        add_five = make_adder(5)
        return add_five(3)

    assert run_and_validate(fn) == 8


def test_pair_in_closure():
    def fn():
        p = Pair(1, 2)

        def get_sum():
            return p.first + p.second

        return get_sum()

    assert run_and_validate(fn) == 3


def test_nested_pair_manipulation():
    def fn():
        def make_pair_adder(p):
            def add_to_pair(x):
                return Pair(p.first + x, p.second + x)

            return add_to_pair

        initial_pair = Pair(1, 2)
        add_to_initial = make_pair_adder(initial_pair)
        result_pair = add_to_initial(3)
        return result_pair.first + result_pair.second

    assert run_and_validate(fn) == 9


def test_bool_returning_closure():
    def fn():
        threshold = 10

        def is_above_threshold(x):
            return x > threshold

        return is_above_threshold(15)

    assert run_and_validate(fn)


def test_arithmetic_in_closure():
    def fn():
        factor = 2.5

        def multiply(x):
            return factor * x

        return multiply(4.0)

    assert run_and_validate(fn) == 10.0


def test_nested_closure_with_record():
    def fn():
        y = 10

        def make_processor(base):
            def process(x):
                return Pair(0, x + y * base)

            return process

        processor = make_processor(10)
        result = processor(15)
        return result.second

    assert run_and_validate(fn)


def test_modify_closure_before_call():
    def fn():
        x = 1
        p = Pair(5, 10)

        def get_sum():
            return x + p.first + p.second

        first_result = get_sum()  # 1 + 5 + 10 = 16
        debug_log(first_result)

        x = 2
        p.first = 7
        p.second = 12

        second_result = get_sum()  # 2 + 7 + 12 = 21
        debug_log(second_result)

        return Array(first_result, second_result)  # [16, 21]

    assert run_and_validate(fn) == Array(16, 21)


def test_return_closure_with_modified_values():
    def fn():
        x = 1
        p = Pair(5, 10)

        def make_getter():
            def get_values():
                return x + p.first + p.second

            return get_values

        getter = make_getter()
        first_result = getter()  # 1 + 5 + 10 = 16

        x = 2
        p.first = 7
        p.second = 12

        second_result = getter()  # 2 + 7 + 12 = 21
        return Array(first_result, second_result)  # [16, 21]

    assert run_and_validate(fn) == Array(16, 21)


def test_pass_closure_after_modification():
    def fn():
        def apply_twice(f):
            return f() + f()

        x = 1
        p = Pair(5, 10)

        def get_sum():
            return x + p.first + p.second

        first_result = apply_twice(get_sum)  # (1+5+10) + (1+5+10) = 32

        x = 2
        p.first = 7
        p.second = 12

        second_result = apply_twice(get_sum)  # (2+7+12) + (2+7+12) = 42
        return Array(first_result, second_result)  # [32, 42]

    assert run_and_validate(fn) == Array(32, 42)


def test_nested_closure_pair_modification():
    def fn():
        p = Pair(1, 2)

        def make_modifier():
            def modify():
                p.first += 1
                p.second += 2
                return p.first + p.second

            return modify

        modifier = make_modifier()
        first = modifier()  # p becomes (2,4), returns 6
        second = modifier()  # p becomes (3,6), returns 9
        return Array(first, second)  # [6, 9]

    assert run_and_validate(fn) == Array(6, 9)


def test_multiple_closures_sharing_state():
    def fn():
        p = Pair(1, 2)

        def make_incrementer():
            def increment():
                p.first += 1
                p.second += 2
                return p.first

            return increment

        def make_getter():
            def get():
                return p.first + p.second

            return get

        incrementer = make_incrementer()
        getter = make_getter()

        first = getter()  # 1 + 2 = 3
        incrementer()  # p becomes (2,4)
        second = getter()  # 2 + 4 = 6
        incrementer()  # p becomes (3,6)
        third = getter()  # 3 + 6 = 9

        return Array(first, second, third)  # [3, 6, 9]

    assert run_and_validate(fn) == Array(3, 6, 9)


def test_higher_order_with_modified_closure():
    def fn():
        def apply_and_sum(f1, f2):
            return f1() + f2()

        x = 1
        y = 2

        def make_first():
            return lambda: x + y

        def make_second():
            return lambda: x * y

        f1 = make_first()
        f2 = make_second()

        first_result = apply_and_sum(f1, f2)  # (1+2) + (1*2) = 5

        x = 2
        y = 3

        second_result = apply_and_sum(f1, f2)  # (2+3) + (2*3) = 11

        return Array(first_result, second_result)  # [5, 11]

    assert run_and_validate(fn) == Array(5, 11)


def test_closure_with_pair_default():
    def fn():
        base = Pair(1, 2)

        def make_adder():
            def add(p=base):
                return p.first + p.second

            return add

        adder = make_adder()
        first = adder()  # 1 + 2 = 3

        base.first = 3
        base.second = 4

        second = adder()  # 3 + 4 = 7
        return Array(first, second)  # [3, 7]

    assert run_and_validate(fn) == Array(3, 7)


def test_closure_with_function_default():
    def fn():
        x = 1

        def default_func():
            return x * 2

        def make_processor():
            def process(f=default_func):
                return f() + x

            return process

        processor = make_processor()
        first = processor()  # (1*2) + 1 = 3

        x = 2
        second = processor()  # (2*2) + 2 = 6

        return Array(first, second)  # [3, 6]

    assert run_and_validate(fn) == Array(3, 6)


def test_nested_default_args():
    def fn():
        p = Pair(1, 2)

        def make_outer(default_value=10):
            def make_inner(multiplier=2):
                def compute(pair=p):
                    return (pair.first + pair.second) * multiplier + default_value

                return compute

            return make_inner

        f1 = make_outer()(3)(Pair(2, 3))  # (2+3)*3 + 10 = 25
        f2 = make_outer(20)()(p)  # (1+2)*2 + 20 = 26

        return Array(f1, f2)  # [25, 26]

    assert run_and_validate(fn) == Array(25, 26)


def test_default_args_with_modification():
    def fn():
        p1 = Pair(1, 2)
        p2 = Pair(3, 4)

        def make_calculator(base=p1):
            def calc(other=p2):
                return base.first + base.second + other.first + other.second

            return calc

        calculator = make_calculator()
        first = calculator()  # 1 + 2 + 3 + 4 = 10

        p1.first = 5
        p2.second = 6

        second = calculator()  # 5 + 2 + 3 + 6 = 16
        return Array(first, second)  # [10, 16]

    assert run_and_validate(fn) == Array(10, 16)


def test_default_function_modification():
    def fn():
        x = 1

        def initial_func():
            return x

        def make_processor(f=initial_func):
            def process(multiplier=2):
                return f() * multiplier

            return process

        processor = make_processor()
        first = processor()  # 1 * 2 = 2

        x = 3
        second = processor(3)  # 3 * 3 = 9

        return Array(first, second)  # [2, 9]

    assert run_and_validate(fn) == Array(2, 9)


def test_mixed_defaults_and_closure():
    def fn():
        base = Pair(1, 2)
        factor = 2

        def make_computer():
            def default_func(p):
                return p.first + p.second

            def compute(f=default_func, p=base, multiplier=factor):
                return f(p) * multiplier

            return compute

        computer = make_computer()
        first = computer()  # (1+2) * 2 = 6

        base.first = 3
        factor = 3  # This won't affect the default arg!

        second = computer(multiplier=4)  # (3+2) * 4 = 20
        return Array(first, second)  # [6, 20]

    assert run_and_validate(fn) == Array(6, 20)


def test_chain_of_defaults():
    def fn():
        p = Pair(1, 2)

        def make_base(pair=p):
            return pair.first + pair.second

        def make_multiplier(f=make_base):
            def multiply(factor=2):
                return f() * factor

            return multiply

        multiplier = make_multiplier()
        first = multiplier()  # (1+2) * 2 = 6

        p.first = 3
        second = multiplier(3)  # (3+2) * 3 = 15

        return Array(first, second)  # [6, 15]

    assert run_and_validate(fn) == Array(6, 15)


def test_nested_pair_default_modification():
    def fn():
        outer = Pair(1, 2)
        inner = Pair(3, 4)

        def make_outer(p1=outer):
            def make_inner(p2=inner):
                def compute(factor=2):
                    return (p1.first + p1.second + p2.first + p2.second) * factor

                return compute

            return make_inner

        computer = make_outer()(Pair(5, 6))
        first = computer()  # (1+2+5+6) * 2 = 28

        outer.first = 7
        inner.second = 8  # Shouldn't affect result since we passed different Pair

        second = computer(3)  # (7+2+5+6) * 3 = 60
        return Array(first, second)  # [28, 60]

    assert run_and_validate(fn) == Array(28, 60)


def test_changing_values_after_closure_creation():
    def fn():
        def make_func(x):
            def f():
                return x

            x *= 2

            return f

        f1 = make_func(1)
        f2 = make_func(3)

        return Array(f1(), f2())

    assert run_and_validate(fn) == Array(2, 6)


def test_nested_default_args_with_side_effects():
    def fn():
        counter = Pair(0, 0)

        def increment_and_return():
            counter.first += 1
            return counter.first

        def make_func(default_val=increment_and_return()):  # noqa: B008
            def inner(x=default_val):
                return x + increment_and_return()

            return inner

        f = make_func()
        first = f()
        second = f(10)
        return Array(first, second)

    assert run_and_validate(fn) == Array(3, 13)


def test_closure_sharing_between_functions():
    def fn():
        shared = Pair(1, 2)

        def make_modifier():
            def modify():
                shared.first *= 2
                shared.second += 3
                return shared.first + shared.second

            return modify

        def make_observer():
            def observe():
                return shared.first * shared.second

            return observe

        modifier = make_modifier()  # shared becomes (2, 5)
        observer = make_observer()  # shared is still (2, 5)

        return Array(modifier(), observer())

    assert run_and_validate(fn) == Array(7, 10)


def test_nested_function_redefinition():
    def fn():
        x = 1

        def make_outer():
            y = 2

            def make_inner():
                z = 3

                def inner():
                    return x + y + z

                z = 4
                return inner

            y = 3
            return make_inner()

        func = make_outer()
        first = func()  # 1 + 3 + 4 = 8
        x = 5
        second = func()  # 5 + 3 + 4 = 12

        return Array(first, second)  # [8, 12]

    assert run_and_validate(fn) == Array(8, 12)


def test_closure_with_conditional_modification():
    def fn():
        p = Pair(1, 2)

        def make_modifier(condition):
            if condition:
                p.first += 5

            def modify():
                if condition:
                    return p.first * 2
                else:
                    return p.second * 3

            if condition:
                p.second += 3
            return modify

        f1 = make_modifier(True)  # p becomes (6, 5)
        f2 = make_modifier(False)  # p unchanged

        return Array(f1(), f2())

    assert run_and_validate(fn) == Array(12, 15)


def test_mutable_default_args():
    def fn():
        def make_adder(pair=Pair(1, 2)):  # noqa: B008
            pair.first += 1
            pair.second += 2

            def add(new_pair=pair):  # Mutable default that gets modified
                new_pair.first += 1
                new_pair.second += 2
                return new_pair.first + new_pair.second

            return add

        adder = make_adder()  # pair becomes (2, 4)
        first = adder()  # pair becomes (3, 6), returns 9
        second = adder()  # pair becomes (4, 8), returns 12

        new_adder = make_adder()  # new pair becomes (5, 10)
        third = new_adder()  # new pair becomes (6, 12), returns 18

        return Array(first, second, third)

    assert run_and_validate(fn) == Array(9, 12, 18)


def test_nested_mutable_defaults():
    def fn():
        def outer(p1=Pair(1, 2)):  # noqa: B008
            p1.first += 1  # p1 becomes (2, 2)

            def inner(p2=p1):  # p2 references the same pair as p1
                def deepest(p3=p2):  # p3 references the same pair as p1 and p2
                    p3.second += 1
                    return p1.first * p2.second  # Same as p3.first * p3.second

                return deepest

            return inner()

        f1 = outer()  # Creates pair (2, 2)
        r1 = f1()  # Modifies to (2, 3), returns 6
        r2 = f1()  # Modifies to (2, 4), returns 8

        f2 = outer()  # Modifies to (3, 4)
        r3 = f2()  # Modifies to (3, 5), returns 15

        return Array(r1, r2, r3)

    assert run_and_validate(fn) == Array(6, 8, 15)


def test_mixed_mutable_immutable_defaults():
    def fn():
        def make_processor(immutable=1, mutable=Pair(1, 2)):  # noqa: B008
            def process(x=immutable, pair=mutable):
                pair.first += x
                pair.second += x
                return pair.first + pair.second

            return process

        proc = make_processor(2)  # immutable=2, creates mutable Pair(1, 2)
        first = proc()  # Pair becomes (3, 4), returns 7
        second = proc()  # Pair becomes (5, 6), returns 11
        third = proc(3)  # Pair becomes (8, 9), returns 17

        return Array(first, second, third)

    assert run_and_validate(fn) == Array(7, 11, 17)


def test_simple_decorator():
    def multiply_result(f):
        def wrapper():
            return f() * 2

        return wrapper

    def fn():
        @multiply_result
        def compute():
            return 5

        return compute()

    assert run_and_validate(fn) == 10


def test_decorator_with_arguments():
    def multiply_by(factor):
        def decorator(f):
            def wrapper():
                return f() * factor

            return wrapper

        return decorator

    def fn():
        @multiply_by(3)
        def compute():
            return 5

        return compute()

    assert run_and_validate(fn) == 15


def test_decorator_with_state():
    def fn():
        counter = Pair(0, 0)  # Using first for call count, second for sum

        def track_calls(f):
            def wrapper(*args):
                counter.first += 1
                result = f(*args)
                counter.second += result
                return result

            return wrapper

        @track_calls
        def add(x, y):
            return x + y

        first = add(2, 3)  # counter becomes (1, 5)
        second = add(4, 5)  # counter becomes (2, 14)

        return Array(first, second, counter.first, counter.second)

    assert run_and_validate(fn) == Array(5, 9, 2, 14)


def test_multiple_decorators():
    def fn():
        def double_result(f):
            def wrapper():
                return f() * 2

            return wrapper

        def add_five(f):
            def wrapper():
                return f() + 5

            return wrapper

        @double_result
        @add_five
        def compute():
            return 3

        return compute()  # (3 + 5) * 2 = 16

    assert run_and_validate(fn) == 16


def test_decorator_with_closure():
    def fn():
        multiplier = 2

        def multiply_by_closure(f):
            def wrapper():
                return f() * multiplier

            return wrapper

        @multiply_by_closure
        def compute():
            return 5

        first = compute()  # 5 * 2 = 10
        multiplier = 3
        second = compute()  # 5 * 3 = 15

        return Array(first, second)

    assert run_and_validate(fn) == Array(10, 15)


def test_decorator_with_mutable_state():
    def fn():
        state = Pair(1, 2)

        def modify_state(f):
            def wrapper():
                state.first += 1
                result = f()
                state.second += 2
                return result

            return wrapper

        @modify_state
        def get_sum():
            return state.first + state.second

        first = get_sum()  # state becomes (2, 2) then (2, 4), returns 4
        second = get_sum()  # state becomes (3, 4) then (3, 6), returns 7

        return Array(first, second)

    assert run_and_validate(fn) == Array(4, 7)


def test_decorator_factory_with_defaults():
    def fn():
        def create_multiplier(factor=2, offset=Pair(0, 0)):  # noqa: B008
            def decorator(f):
                def wrapper():
                    offset.first += 1
                    result = f() * factor + offset.first
                    return result

                return wrapper

            return decorator

        @create_multiplier()
        def compute1():
            return 5

        @create_multiplier(factor=3)
        def compute2():
            return 5

        first = compute1()  # 5 * 2 + 1 = 11
        second = compute1()  # 5 * 2 + 2 = 12
        third = compute2()  # 5 * 3 + 3 = 18

        return Array(first, second, third)

    assert run_and_validate(fn) == Array(11, 12, 18)


def test_nested_decorator_with_shared_state():
    def fn():
        shared = Pair(1, 2)

        def outer_decorator(f):
            def outer_wrapper():
                shared.first *= 2

                @inner_decorator
                def inner():
                    return f() + shared.first

                return inner()

            return outer_wrapper

        def inner_decorator(f):
            def inner_wrapper():
                shared.second *= 2
                return f() + shared.second

            return inner_wrapper

        @outer_decorator
        def compute():
            return 1

        first = compute()  # shared becomes (2, 4), returns 1 + 2 + 4 = 7
        second = compute()  # shared becomes (4, 8), returns 1 + 4 + 8 = 13

        return Array(first, second)

    assert run_and_validate(fn) == Array(7, 13)


def test_decorator_with_conditional_behavior():
    def fn():
        state = Pair(0, 0)

        def conditional_decorator(condition):
            def decorator(f):
                def wrapper():
                    if condition:
                        state.first += 1
                        return f() * 2
                    else:
                        state.second += 1
                        return f() * 3

                return wrapper

            return decorator

        @conditional_decorator(True)
        def compute1():
            return 5

        @conditional_decorator(False)
        def compute2():
            return 5

        return Array(compute1(), compute2(), state.first, state.second)

    assert run_and_validate(fn) == Array(10, 15, 1, 1)


def test_decorator_preserving_default_args():
    def fn():
        def preserve_defaults(f):
            def wrapper(x=None, pair=None):
                if x is None:
                    x = 1
                if pair is None:
                    pair = Pair(1, 2)
                pair.first += x
                return f(x=x, pair=pair)

            return wrapper

        @preserve_defaults
        def compute(x=1, pair=Pair(1, 2)):  # noqa: B008
            return pair.first + x

        first = compute()  # Uses defaults, pair becomes (2, 2), returns 3
        second = compute(2)  # New pair (3, 2), returns 5
        third = compute(pair=Pair(5, 5))  # Pair becomes (6, 5), returns 7

        return Array(first, second, third)

    assert run_and_validate(fn) == Array(3, 5, 7)


def multiply(func=None, factor=2):
    if func is None:

        def decorator(f):
            def wrapper(*args, **kwargs):
                return f(*args, **kwargs) * factor

            return wrapper

        return decorator

    def wrapper(*args, **kwargs):
        return func(*args, **kwargs) * factor

    return wrapper


def test_external_flexible_decorator():
    def fn():
        @multiply
        def compute1():
            return 5

        @multiply(factor=3)
        def compute2():
            return 5

        return Array(compute1(), compute2())  # [10, 15]

    assert run_and_validate(fn) == Array(10, 15)


def test_internal_flexible_decorator():
    def fn():
        state = Pair(1, 0)  # first: multiplier, second: call count

        def counter(func=None, factor=None):
            if func is None:

                def decorator(f):
                    return counter(f, factor)

                return decorator

            def wrapper(*args, **kwargs):
                actual_factor = state.first if factor is None else factor
                state.second += 1
                return func(*args, **kwargs) * actual_factor

            return wrapper

        @counter
        def compute1():
            return 5

        @counter(factor=3)
        def compute2():
            return 5

        r1 = compute1()  # 5 * 1 = 5
        r2 = compute2()  # 5 * 3 = 15
        calls1 = state.second  # 2 calls

        # Modify state
        state.first = 2

        r3 = compute1()  # 5 * 2 = 10
        r4 = compute2()  # 5 * 3 = 15
        calls2 = state.second  # 4 calls

        return Array(r1, r2, calls1, r3, r4, calls2)

    assert run_and_validate(fn) == Array(5, 15, 2, 10, 15, 4)


def test_flexible_decorator_with_mutable_state():
    def fn():
        def accumulate(func=None, storage=None):
            actual_storage = Pair(0, 0) if storage is None else storage

            if func is None:

                def decorator(f):
                    def wrapper(*args, **kwargs):
                        actual_storage.first += 1
                        result = f(*args, **kwargs)
                        actual_storage.second += result
                        return result

                    return wrapper

                return decorator

            def wrapper(*args, **kwargs):
                actual_storage.first += 1
                result = func(*args, **kwargs)
                actual_storage.second += result
                return result

            return wrapper

        shared_store = Pair(0, 0)  # first: call count, second: sum

        @accumulate(storage=shared_store)
        def compute1():
            return 5

        @accumulate()  # Uses its own Pair
        def compute2():
            return 3

        r1 = compute1()  # 5, shared_store becomes (1, 5)
        r2 = compute2()  # 3, different store becomes (1, 3)
        r3 = compute1()  # 5, shared_store becomes (2, 10)
        calls = shared_store.first  # 2 calls to compute1
        total = shared_store.second  # Sum of 10 from compute1

        return Array(r1, r2, r3, calls, total)

    assert run_and_validate(fn) == Array(5, 3, 5, 2, 10)


def test_nested_flexible_decorators():
    def fn():
        def multiply(func=None, factor=2):
            if func is None:

                def decorator(f):
                    def wrapper(*args, **kwargs):
                        return f(*args, **kwargs) * factor

                    return wrapper

                return decorator

            def wrapper(*args, **kwargs):
                return func(*args, **kwargs) * factor

            return wrapper

        def add(func=None, value=1):
            if func is None:

                def decorator(f):
                    def wrapper(*args, **kwargs):
                        return f(*args, **kwargs) + value

                    return wrapper

                return decorator

            def wrapper(*args, **kwargs):
                return func(*args, **kwargs) + value

            return wrapper

        @multiply
        @add(value=2)
        def compute1():
            return 5  # (5 + 2) * 2 = 14

        @multiply(factor=3)
        @add
        def compute2():
            return 5  # (5 + 1) * 3 = 18

        @add
        @multiply(factor=2)
        def compute3():
            return 5  # (5 * 2) + 1 = 11

        return Array(compute1(), compute2(), compute3())

    assert run_and_validate(fn) == Array(14, 18, 11)


def test_flexible_decorator_with_dynamic_state():
    def fn():
        state = Pair(1, 1)  # first: base value, second: current multiplier

        def dynamic(func=None, capture_state=False):
            if func is None:

                def decorator(f):
                    # Capture current state if requested
                    base = state.first if capture_state else None

                    def wrapper(*args, **kwargs):
                        multiplier = state.second
                        if base is not None:
                            return f(*args, **kwargs) * multiplier + base
                        return f(*args, **kwargs) * multiplier

                    return wrapper

                return decorator

            def wrapper(*args, **kwargs):
                return func(*args, **kwargs) * state.second

            return wrapper

        @dynamic
        def compute1():
            return 5

        @dynamic(capture_state=True)
        def compute2():
            return 5

        r1 = compute1()  # 5 * 1 = 5
        r2 = compute2()  # 5 * 1 + 1 = 6

        # Modify state
        state.first = 2
        state.second = 3

        r3 = compute1()  # 5 * 3 = 15
        r4 = compute2()  # 5 * 3 + 1 = 16 (uses captured base=1)

        return Array(r1, r2, r3, r4)

    assert run_and_validate(fn) == Array(5, 6, 15, 16)


def test_nested_function_with_complex_signature():
    def fn():
        def nested(pos_only_1, pos_only_2, /, pos_or_kw_1, pos_or_kw_2, *args, kw_only_1, kw_only_2=3, **kwargs):
            return pos_only_1 + pos_only_2 + pos_or_kw_1 + pos_or_kw_2 + kw_only_1 + kw_only_2

        return nested(1, 2, 3, 4, 12345, kw_only_1=5, kw_only_2=6, extra=7)

    assert run_and_validate(fn) == 21



from sonolus.script.array import Array
from sonolus.script.containers import VarArray
from sonolus.script.debug import debug_log
from sonolus.script.iterator import maybe_next
from sonolus.script.maybe import Nothing, Some
from tests.script.conftest import run_and_validate


def test_simple_some():
    def fn():
        def f():
            return Some(1)

        result = f()
        if result.is_some:
            return result.get()
        else:
            return 0

    assert run_and_validate(fn) == 1


def test_simple_nothing():
    def fn():
        def f():
            return Nothing

        result = f()
        if result.is_some:
            return 1
        else:
            return 0

    assert run_and_validate(fn) == 0


def test_test_multiple_maybe_returns():
    def fn():
        results = VarArray[int, 10].new()

        def f(x):
            if x > 5:
                return Nothing
            if x <= 2:
                return Nothing
            return Some(x * 2)

        for i in range(10):
            result = f(i)
            if result.is_some:
                results.append(result.get())

        return results

    assert list(run_and_validate(fn)) == [6, 8, 10]


def test_maybe_in_generator():
    def fn():
        def generator(n):
            for i in range(n):
                if i % 2 == 0:
                    yield Some(i)
                else:
                    yield Nothing

        results = VarArray[int, 10].new()
        for result in generator(10):
            if result.is_some:
                results.append(result.get())

        return results

    assert list(run_and_validate(fn)) == [0, 2, 4, 6, 8]


def test_maybe_map_with_some():
    def fn():
        maybe_val = Some(5)
        mapped = maybe_val.map(lambda x: x * 2)
        if mapped.is_some:
            return mapped.get()
        else:
            return 0

    assert run_and_validate(fn) == 10


def test_maybe_map_with_nothing():
    def fn():
        maybe_val = Nothing
        mapped = maybe_val.map(lambda x: x * 2)
        if mapped.is_some:
            return 1
        else:
            return 0

    assert run_and_validate(fn) == 0


def test_maybe_flat_map_with_some():
    def fn():
        maybe_val = Some(5)
        flat_mapped = maybe_val.flat_map(lambda x: Some(x * 3) if x > 0 else Nothing)
        if flat_mapped.is_some:
            return flat_mapped.get()
        else:
            return 0

    assert run_and_validate(fn) == 15


def test_maybe_flat_map_with_nothing():
    def fn():
        maybe_val = Nothing
        flat_mapped = maybe_val.flat_map(lambda x: Some(x * 3) if x > 0 else Nothing)
        if flat_mapped.is_some:
            return 1
        else:
            return 0

    assert run_and_validate(fn) == 0


def test_maybe_flat_map_returns_nothing():
    def fn():
        maybe_val = Some(-5)
        flat_mapped = maybe_val.flat_map(lambda x: Nothing if x < 0 else Some(x))
        if flat_mapped.is_some:
            return 1
        else:
            return 0

    assert run_and_validate(fn) == 0


def test_maybe_or_default_with_some():
    def fn():
        maybe_val = Some(42)
        return maybe_val.or_default(10)

    assert run_and_validate(fn) == 42


def test_maybe_or_default_with_nothing():
    def fn():
        maybe_val = Nothing
        return maybe_val.or_default(10)

    assert run_and_validate(fn) == 10


def test_maybe_or_else_with_some():
    def fn():
        maybe_val = Some(42)
        return maybe_val.or_else(lambda: 10)

    assert run_and_validate(fn) == 42


def test_maybe_or_else_with_nothing():
    def fn():
        maybe_val = Nothing
        return maybe_val.or_else(lambda: 15)

    assert run_and_validate(fn) == 15


def test_maybe_or_else_dynamic():
    def fn():
        iterator = iter(Array(1))
        debug_log(maybe_next(iterator).or_else(lambda: 4))
        debug_log(maybe_next(iterator).or_else(lambda: 5))
        debug_log(maybe_next(iterator).or_else(lambda: 6))

    run_and_validate(fn)


def test_maybe_tuple_with_some():
    def fn():
        maybe_val = Some(42)
        is_present, value = maybe_val.tuple
        if is_present:
            return value
        else:
            return 0

    assert run_and_validate(fn) == 42


def test_maybe_tuple_with_nothing():
    def fn():
        maybe_val = Nothing
        is_present, value = maybe_val.tuple
        if is_present:
            return value
        else:
            return 123

    assert run_and_validate(fn) == 123


def test_maybe_is_nothing_property():
    def fn():
        some_val = Some(42)
        nothing_val = Nothing

        result = 0
        if some_val.is_nothing:
            result += 1
        if nothing_val.is_nothing:
            result += 10

        return result

    assert run_and_validate(fn) == 10


def test_maybe_next_with_empty_iterable():
    def fn():
        empty_list = Array[int, 0]()
        result = maybe_next(iter(empty_list))
        if result.is_some:
            return 1
        else:
            return 0

    assert run_and_validate(fn) == 0


def test_maybe_next_with_non_empty_iterable():
    def fn():
        items = Array(1, 2, 3)
        result = maybe_next(iter(items))
        if result.is_some:
            return result.get()
        else:
            return 0

    assert run_and_validate(fn) == 1



from hypothesis import given
from hypothesis import strategies as st

from sonolus.build.dev_server import (
    DecodeCommand,
    ExitCommand,
    HelpCommand,
    RebuildCommand,
    parse_dev_command,
)


def test_parse_rebuild_command_full():
    result = parse_dev_command("rebuild")
    assert isinstance(result, RebuildCommand)


def test_parse_rebuild_command_alias():
    result = parse_dev_command("r")
    assert isinstance(result, RebuildCommand)


def test_parse_rebuild_command_with_extra_args_invalid():
    result = parse_dev_command("rebuild extra")
    assert result is None


def test_parse_rebuild_command_with_extra_args_invalid_alias():
    result = parse_dev_command("r extra")
    assert result is None


def test_parse_decode_command_full():
    result = parse_dev_command("decode 42")
    assert isinstance(result, DecodeCommand)
    assert result.message_code == 42


def test_parse_decode_command_alias():
    result = parse_dev_command("d 123")
    assert isinstance(result, DecodeCommand)
    assert result.message_code == 123


def test_parse_decode_command_zero():
    result = parse_dev_command("decode 0")
    assert isinstance(result, DecodeCommand)
    assert result.message_code == 0


def test_parse_decode_command_negative():
    result = parse_dev_command("decode -5")
    assert isinstance(result, DecodeCommand)
    assert result.message_code == -5


def test_parse_decode_command_missing_arg():
    result = parse_dev_command("decode")
    assert result is None


def test_parse_decode_command_missing_arg_alias():
    result = parse_dev_command("d")
    assert result is None


def test_parse_decode_command_non_int_arg():
    result = parse_dev_command("decode abc")
    assert result is None


def test_parse_decode_command_non_int_arg_alias():
    result = parse_dev_command("d xyz")
    assert result is None


def test_parse_decode_command_extra_args():
    result = parse_dev_command("decode 42 extra")
    assert result is None


def test_parse_help_command_full():
    result = parse_dev_command("help")
    assert isinstance(result, HelpCommand)


def test_parse_help_command_alias():
    result = parse_dev_command("h")
    assert isinstance(result, HelpCommand)


def test_parse_help_command_with_extra_args_invalid():
    result = parse_dev_command("help extra")
    assert result is None


def test_parse_help_command_with_extra_args_invalid_alias():
    result = parse_dev_command("h extra")
    assert result is None


def test_parse_quit_command_full():
    result = parse_dev_command("quit")
    assert isinstance(result, ExitCommand)


def test_parse_quit_command_alias():
    result = parse_dev_command("q")
    assert isinstance(result, ExitCommand)


def test_parse_quit_command_with_extra_args_invalid():
    result = parse_dev_command("quit extra")
    assert result is None


def test_parse_quit_command_with_extra_args_invalid_alias():
    result = parse_dev_command("q extra")
    assert result is None


def test_parse_unknown_command():
    result = parse_dev_command("unknown")
    assert result is None


def test_parse_empty_command():
    result = parse_dev_command("")
    assert result is None


def test_parse_whitespace_only():
    result = parse_dev_command("   ")
    assert result is None


def test_parse_invalid_shell_syntax_unclosed_quote():
    result = parse_dev_command('decode "unclosed')
    assert result is None


def test_parse_invalid_shell_syntax_unclosed_single_quote():
    result = parse_dev_command("decode 'unclosed")
    assert result is None


@given(st.text())
def test_parse_dev_command_arbitrary_strings(command_line):
    # Should never throw
    result = parse_dev_command(command_line)
    assert result is None or hasattr(result, "execute")



"""Test cases intended to cover more complex control flow."""

from enum import IntEnum

from sonolus.script.array import Array
from sonolus.script.containers import VarArray
from sonolus.script.debug import debug_log
from sonolus.script.num import Num
from sonolus.script.record import Record
from tests.script.conftest import run_and_validate
from tests.script.test_record import Pair


class Pair2[T, U](Record):
    first: T
    second: U


class Pair3[T, U](Record):
    first: T
    second: U


class Color(IntEnum):
    RED = 1
    GREEN = 2
    BLUE = 3


class Point(Record):
    x: Num
    y: Num


class Circle(Record):
    center: Point
    radius: Num


class Rectangle(Record):
    top_left: Point
    bottom_right: Point


class Shape[Kind, Data](Record):
    kind: Kind
    data: Data


def test_match_pair_and_num_checking():
    a = Pair(1, 2)
    b = 3
    c = Pair2(4, 5)
    d = "hello"

    def m(x):
        match x:
            case Pair(a, b):
                debug_log(a)
                debug_log(b)
            case Num(n):
                debug_log(n)
            case Pair2(a, b):
                debug_log(a + b)
                debug_log(a - b)
            case _:
                debug_log(-1)
        debug_log(123)

    def fn():
        m(a)
        m(b)
        m(c)
        m(d)

    run_and_validate(fn)


def test_match_shapes_and_collections_patterns():
    p1 = Point(1, 2)
    p2 = Point(3, 4)
    c = Circle(p1, 5)
    r = Rectangle(p1, p2)
    color = Color.RED
    num = 42
    tpl = (1, 2, 3)
    text = "test"

    def m(x):
        match x:
            case Point(a, b):
                debug_log(a + b)
                debug_log(a * b)
            case Circle(center=Point(x=Num(a), y=Num(b)), radius=Num(r)):
                debug_log(a)
                debug_log(b)
                debug_log(r)
            case Rectangle(
                top_left=Point(x=Num(a), y=Num(b)),
                bottom_right=Point(x=Num(c), y=Num(d)),
            ):
                debug_log(a + c)
                debug_log(b + d)
            case Color.RED | Color.BLUE:
                debug_log(x)
            case Num(n):
                debug_log(n * 2)
            case (Num(a), Num(b), Num(c)):
                debug_log(a + b + c)
            case _:
                debug_log(-1)
        debug_log(123)

    def fn():
        m(p1)
        m(c)
        m(r)
        m(color)
        m(num)
        m(tpl)
        m(text)

    run_and_validate(fn)


def test_match_nested_conditions():
    p = Point(5, 5)
    c = Circle(Point(0, 0), 15)
    r = Rectangle(Point(0, 0), Point(20, 20))
    n = 100
    s = "hello"

    def m(x):
        match x:
            case Circle(center=Point(x=Num(a), y=Num(b)), radius=Num(r)) if r > 10:
                debug_log(a)
                debug_log(b)
                debug_log(r)
            case Rectangle(
                top_left=Point(x=Num(a), y=Num(b)),
                bottom_right=Point(x=Num(c), y=Num(d)),
            ):
                area = (c - a) * (d - b)
                debug_log(area)
            case Point(x=Num(a), y=Num(b)) if a == b:
                debug_log(a)
            case Num(n) if n > 50:
                debug_log(n)
            case _:
                debug_log(-2)
        debug_log(123)

    def fn():
        m(c)
        m(r)
        m(p)
        m(n)
        m(s)

    run_and_validate(fn)


def test_match_nesting_tuples():
    t1 = (1, 2, 3)
    t2 = ((5, 6), 7, 8)
    t3 = ((10, 11), 7, 8)
    t4 = (1, 2, 3, 4)
    t5 = (1, 2)
    t6 = (1, 2, 3, 4, 5)
    t7 = (4, 3, 2, 1)

    def m(x):
        match x:
            case (Num() as a, b, c):
                debug_log(a)
                debug_log(b)
                debug_log(c)
            case ((a, b) as x, c, d) as y:
                debug_log(a)
                debug_log(b)
                debug_log(c)
                debug_log(d)
                for i in x:
                    for j in y:
                        if isinstance(j, tuple):
                            for k in j:
                                debug_log(k)
                        else:
                            debug_log(i + j)
            case (10, 11, 7, 8):
                debug_log(10)
            case (a, b):
                debug_log(a)
                debug_log(b)
            case (a, b, c, d):
                debug_log(a)
                debug_log(b)
                debug_log(c)
                debug_log(d)
            case _:
                debug_log(-1)
        debug_log(123)

    def fn():
        m(t1)
        m(t2)
        m(t3)
        m(t4)
        m(t5)
        m(t6)
        m(t7)

    run_and_validate(fn)


def test_match_nest_arrays():
    a1 = Array(1, 2, 3)
    a2 = Array(Array(1, 2), Array(3, 4), Array(5, 6))
    a3 = Array(1, 2, 4)
    a4 = Array(1, 2, 3, 4)
    a5 = Array(1, 2)
    a6 = Array(1, 2, 3, 4, 5)
    a7 = Array(Array(4, 3), Array(2, 1))
    a8 = Array(Array(Array(123)))
    a9 = Array(Array(Array(Array(123))))
    a10 = Array(Array(Array(Array(456))))
    a11 = 1

    def m(x):
        match x:
            case (Num() as a, b, c):
                debug_log(a)
                debug_log(b)
                debug_log(c)
            case ((a, b) as x, (c, d) as y):
                debug_log(a)
                debug_log(b)
                debug_log(c)
                debug_log(d)
                for i in x:
                    for j in y:
                        if isinstance(j, Array):
                            for k in j:
                                debug_log(k)
                        else:
                            debug_log(i + j)
            case (1, 2, 3):
                debug_log(1)
            case (a, b) if isinstance(a, Num):
                debug_log(a)
                debug_log(b)
            case (a, b, c, d):
                debug_log(a)
                debug_log(b)
                debug_log(c)
                debug_log(d)
            case ((((123,),),),):
                debug_log(111)
            case ((((v,),),),):
                debug_log(v)
            case Array() as a:
                for i in a:
                    if isinstance(i, Num):
                        debug_log(10 * i)
                    elif isinstance(i, Array):
                        for j in i:
                            if isinstance(j, Num):
                                debug_log(9 * j)
        debug_log(123)

    def fn():
        m(a1)
        m(a2)
        m(a3)
        m(a4)
        m(a5)
        m(a6)
        m(a7)
        m(a8)
        m(a9)
        m(a10)
        m(a11)

    run_and_validate(fn)


def test_match_nest_var_arrays():
    def m(x):
        match x:
            case (Num() as a, b, c):
                debug_log(a)
                debug_log(b)
                debug_log(c)
            case ((a, b) as x, (c, d) as y):
                debug_log(a)
                debug_log(b)
                debug_log(c)
                debug_log(d)
                for i in x:
                    for j in y:
                        if isinstance(j, Array):
                            for k in j:
                                debug_log(k)
                        else:
                            debug_log(i + j)
            case (1, 2, 3):
                debug_log(1)
            case (a, b) if isinstance(a, Num):
                debug_log(a)
                debug_log(b)
            case (a, b, c, d) if isinstance(a, Num):
                debug_log(a)
                debug_log(b)
                debug_log(c)
                debug_log(d)
            case ((((123,),),),):
                debug_log(111)
            case ((((v,),),),):
                debug_log(v)
            case Array() as a:
                for i in a:
                    if isinstance(i, Num):
                        debug_log(10 * i)
                    elif isinstance(i, Array):
                        for j in i:
                            if isinstance(j, Num):
                                debug_log(9 * j)

    def fn():
        a1 = VarArray[Num, 10].new()
        a1.append(1)
        m(a1)
        a1.append(2)
        m(a1)
        a1.append(3)
        m(a1)
        a1.append(4)
        m(a1)
        a1.append(5)
        m(a1)
        a1.clear()
        m(a1)

        a2 = VarArray[VarArray[Num, 2], 3].new()
        a2_0 = VarArray[Num, 2].new()
        a2_0.append(1)
        a2_0.append(2)
        a2.append(a2_0)
        m(a2)
        a2_1 = VarArray[Num, 2].new()
        a2_1.append(3)
        a2_1.append(4)
        a2.append(a2_1)
        m(a2)
        a2_2 = VarArray[Num, 2].new()
        a2_2.append(5)
        a2_2.append(6)
        a2.append(a2_2)
        m(a2)

        a3 = VarArray[Array[Array[Array[int, 1], 1], 1], 1].new()
        m(a3)
        a3_0 = Array(Array(Array(1)))
        a3.append(a3_0)
        m(a3)
        a3.clear()
        m(a3)

    run_and_validate(fn)


def test_match_nested_shape_records():
    circle = Shape(kind="circle", data=Circle(center=Point(0, 0), radius=10))
    rectangle = Shape(
        kind="rectangle",
        data=Rectangle(top_left=Point(0, 0), bottom_right=Point(20, 10)),
    )
    unknown = Shape(kind="unknown", data=None)
    num = 5

    def m(x):
        match x:
            case Shape(
                kind="circle",
                data=Circle(center=Point(x=Num(a), y=Num(b)), radius=Num(r)),
            ):
                debug_log(a)
                debug_log(b)
                debug_log(r)
            case Shape(
                kind="rectangle",
                data=Rectangle(
                    top_left=Point(x=Num(a), y=Num(b)),
                    bottom_right=Point(x=Num(c), y=Num(d)),
                ),
            ):
                debug_log((c - a) * (d - b))
            case Shape(kind=_, data=None):
                debug_log(-3)
            case Num(n):
                debug_log(n * n)
            case _:
                debug_log(-4)
        debug_log(123)

    def fn():
        m(circle)
        m(rectangle)
        m(unknown)
        m(num)
        m("test")

    run_and_validate(fn)


def test_match_enum_patterns():
    color1 = Color.RED
    color2 = Color.GREEN
    color3 = Color.BLUE
    num = 7
    text = "color"

    def m(x):
        match x:
            case Num(Color.RED | Color.BLUE):
                debug_log(x)
            case Color.GREEN:
                debug_log(100)
            case Num(n):
                debug_log(n + 10)
        debug_log(123)

    def fn():
        m(color1)
        m(color2)
        m(color3)
        m(num)
        m(text)

    run_and_validate(fn)


def test_match_nested_shapes_and_pair():
    p1 = Point(1, 1)
    p2 = Point(2, 2)
    p3 = Point(3, 3)
    c = Circle(center=p1, radius=10)
    r = Rectangle(top_left=p2, bottom_right=p3)
    shape_circle = Shape(kind="circle", data=c)
    shape_rectangle = Shape(kind="rectangle", data=r)
    pair = Pair(first=p1, second=r)

    def m(x):
        match x:
            case Shape(
                kind="circle",
                data=Circle(center=Point(x=Num(a), y=Num(b)), radius=Num(r)),
            ) if r > 5:
                debug_log(a + b)
                debug_log(r)
            case Shape(
                kind="rectangle",
                data=Rectangle(
                    top_left=Point(x=Num(a), y=Num(b)),
                    bottom_right=Point(x=Num(c), y=Num(d)),
                ),
            ) if (c - a) > 0 and (d - b) > 0:
                debug_log((c - a) * (d - b))
            case Pair(first=Point(x=Num(a), y=Num(b)) as p, second=_):
                debug_log(p.x + p.y)
                debug_log(a * b)
            case Point(x=Num(a), y=Num(b)) if a == b:
                debug_log(a * b)
            case _:
                pass
        debug_log(123)

    def fn():
        m(shape_circle)
        m(shape_rectangle)
        m(pair)
        m(p1)
        m(p2)
        m(p3)

    run_and_validate(fn)


def test_match_partial_binding():
    p1 = Pair(1, 2)
    p2 = Pair2(3, 4)
    p3 = Pair(5, 6)
    p4 = Pair2(7, 8)
    p5 = Pair(3, 2)
    p6 = Pair2(1, 3)
    p7 = Pair(Pair(1, 2), Pair(3, 4))
    p8 = Pair2(Pair2(1, 2), Pair2(3, 4))
    p9 = Pair3(1, 2)

    def m(x):
        match x:
            case Pair(1, _) | Pair2(_, _) if isinstance(x, Pair3):
                # This is unreachable, but it's a valid pattern we're testing
                debug_log(123)
            case Pair(1, b) | Pair2(2, b):
                debug_log(b)
            case Pair(a, 2) | Pair2(a, 3):
                debug_log(a)
            case Pair(Num(a), b) | Pair2(a, Num(b)) if a > 3:
                debug_log(a + b)
            case Pair(a, b) | Pair2(a, b) if isinstance(a, Num) and a <= 3:
                debug_log(a - b)
            case Pair(Pair() as a) | Pair2(7, a):
                if isinstance(a, Pair):
                    debug_log(a.first)
                    debug_log(a.second)
                else:
                    debug_log(a)
            case _:
                debug_log(-1)

    def fn():
        m(p1)
        m(p2)
        m(p3)
        m(p4)
        m(p5)
        m(p6)
        m(p7)
        m(p8)
        m(p9)

    run_and_validate(fn)



from __future__ import annotations

from collections.abc import Iterable

from sonolus.script.debug import assert_true, is_static_true
from sonolus.script.internal.builtin_impls import _max
from sonolus.script.internal.context import ctx
from sonolus.script.internal.impl import validate_value
from sonolus.script.internal.math_impls import _ceil, _round
from sonolus.script.internal.meta_fn import meta_fn, perf_meta_fn
from sonolus.script.internal.tuple_impl import TupleImpl
from sonolus.script.interval import clamp
from sonolus.script.num import Num, _is_num
from sonolus.script.record import Record

enable_np = False


@meta_fn
def _validate_num(value: float):
    if enable_np:
        try:
            import numpy as np
        except ImportError:
            np = None
        if np is not None and isinstance(value, np.float32):
            return value
    value = validate_value(value)
    if not _is_num(value):
        raise TypeError("Only numeric arguments to float() are supported")
    return value


class _UInt32(Record):
    hi: int
    lo: int

    @classmethod
    @perf_meta_fn
    def of(cls, value: int):
        lo = value % (2**16)
        hi = (value // (2**16)) % (2**16)
        return _UInt32._(hi, lo)

    @classmethod
    @meta_fn
    def _(cls, hi: int, lo: int) -> _UInt32:
        # This creates read-only instances, which helps with constant folding in the frontend and build times.
        return _UInt32._raw(hi=_validate_num(hi), lo=_validate_num(lo))

    @classmethod
    @perf_meta_fn
    def zero(cls) -> _UInt32:
        return _UInt32._(0, 0)

    @classmethod
    @perf_meta_fn
    def one(cls) -> _UInt32:
        return _UInt32._(0, 1)

    @classmethod
    @perf_meta_fn
    def _carry_add(cls, a: int, b: int) -> tuple[int, int]:
        if is_static_true(a == 0):
            return b, 0
        if is_static_true(b == 0):
            return a, 0

        ab_raw = a + b
        ab = ab_raw % (2**16)
        carry = ab_raw >= (2**16)
        return ab, carry

    @classmethod
    @perf_meta_fn
    def _borrow_sub(cls, a: int, b: int) -> tuple[int, int]:
        if is_static_true(a == b):
            return 0, 0
        if is_static_true(b == 0):
            return a, 0

        ab_raw = a - b
        ab = ab_raw % (2**16)
        borrow = ab_raw < 0
        return ab, borrow

    @classmethod
    @perf_meta_fn
    def _carry_mul(cls, a: int, b: int) -> tuple[int, int]:
        # Have to be careful since 32-bit floats only have 24 bits of precision

        if is_static_true(a == 0) or is_static_true(b == 0):
            return 0, 0
        if is_static_true(a == 1):
            return b, 0
        if is_static_true(b == 1):
            return a, 0

        blo = b % (2**8)
        bhi = b // (2**8)
        ablo_raw = a * blo
        abhi_raw = a * bhi * (2**8)

        ablo_lo = ablo_raw % (2**16)
        ablo_hi = ablo_raw // (2**16)
        abhi_lo = abhi_raw % (2**16)
        abhi_hi = abhi_raw // (2**16)
        lo_raw = ablo_lo + abhi_lo
        lo = lo_raw % (2**16)
        carry = ablo_hi + abhi_hi + (lo_raw >= (2**16))
        return lo, carry

    @classmethod
    @perf_meta_fn
    def _wrap_mul(cls, a: int, b: int) -> int:
        # We have this for the same reason as _carry_mul

        if is_static_true(a == 0) or is_static_true(b == 0):
            return 0
        if is_static_true(a == 1):
            return b
        if is_static_true(b == 1):
            return a

        blo = b % (2**8)
        bhi = b // (2**8)
        ablo_raw = a * blo
        abhi_raw = a * bhi * (2**8)

        ablo_lo = ablo_raw % (2**16)
        abhi_lo = abhi_raw % (2**16)
        lo_raw = ablo_lo + abhi_lo
        lo = lo_raw % (2**16)
        return lo

    @perf_meta_fn
    def __add__(self, other: _UInt32) -> _UInt32:
        if is_static_true(self == _UInt32.zero()):
            return +other
        if is_static_true(other == _UInt32.zero()):
            return +self
        lo, carry = self._carry_add(self.lo, other.lo)
        hi = (self.hi + other.hi + carry) % (2**16)
        return _UInt32._(hi, lo)

    @perf_meta_fn
    def __sub__(self, other: _UInt32) -> _UInt32:
        if is_static_true(other == _UInt32.zero()):
            return +self
        lo, borrow = self._borrow_sub(self.lo, other.lo)
        hi = (self.hi - other.hi - borrow) % (2**16)
        return _UInt32._(hi, lo)

    @perf_meta_fn
    def __mul__(self, other: _UInt32) -> _UInt32:
        if is_static_true(self == _UInt32.one()):
            return +other
        if is_static_true(other == _UInt32.one()):
            return +self
        lo_lo, carry_lo_lo = self._carry_mul(self.lo, other.lo)
        hi_lo = self._wrap_mul(self.hi, other.lo)
        lo_hi = self._wrap_mul(self.lo, other.hi)
        # hi_hi is ignored since it would overflow entirely

        lo = lo_lo
        hi = (hi_lo + lo_hi + carry_lo_lo) % (2**16)
        return _UInt32._(hi, lo)

    @meta_fn
    def __eq__(self, other):
        if ctx():
            return Num.and_(self.hi == other.hi, self.lo == other.lo)
        return (self.hi, self.lo) == (other.hi, other.lo)

    @meta_fn
    def __ne__(self, other):
        if ctx():
            return Num.or_(self.hi != other.hi, self.lo != other.lo)
        return (self.hi, self.lo) != (other.hi, other.lo)

    @meta_fn
    def __lt__(self, other):
        if ctx():
            hi_lt = self.hi < other.hi
            hi_eq = self.hi == other.hi
            lo_lt = self.lo < other.lo
            return Num.or_(hi_lt, Num.and_(hi_eq, lo_lt))
        return (self.hi, self.lo) < (other.hi, other.lo)

    @meta_fn
    def __le__(self, other):
        if ctx():
            hi_lt = self.hi < other.hi
            hi_eq = self.hi == other.hi
            lo_le = self.lo <= other.lo
            return Num.or_(hi_lt, Num.and_(hi_eq, lo_le))
        return (self.hi, self.lo) <= (other.hi, other.lo)

    @meta_fn
    def __gt__(self, other):
        if ctx():
            hi_gt = self.hi > other.hi
            hi_eq = self.hi == other.hi
            lo_gt = self.lo > other.lo
            return Num.or_(hi_gt, Num.and_(hi_eq, lo_gt))
        return (self.hi, self.lo) > (other.hi, other.lo)

    @meta_fn
    def __ge__(self, other):
        if ctx():
            hi_gt = self.hi > other.hi
            hi_eq = self.hi == other.hi
            lo_ge = self.lo >= other.lo
            return Num.or_(hi_gt, Num.and_(hi_eq, lo_ge))
        return (self.hi, self.lo) >= (other.hi, other.lo)

    __hash__ = None


# Technically we could do a bit more and still fit in the number of, distinct finite 32-bit floats,
# but for simplicity, we limit ourselves to 31 bits.
_MAX_TOTAL_STEPS_UINT32 = _UInt32._(2**15, 0)
_HALF_MAX_TOTAL_STEPS_UINT32 = _UInt32._(2**14, 0)
_HALF_MAX_TOTAL_STEPS_UINT32_MINUS_ONE = _UInt32._(2**14 - 1, 2**16 - 1)


@perf_meta_fn
def quantize_to_step(value: float, start: float, stop: float, step: float) -> tuple[int, int]:
    """Quantize a float value by step size within a range and return the step number and total steps in the range.

    If value is between start and stop, start + step_number * step will be approximately equal to value,
    where step_number is the first element of the returned tuple.

    Args:
        value: The float value to quantize.
        start: The start of the range. The range is inclusive of this value.
        stop: The end of the range. The range is exclusive of this value. Must be strictly greater than start.
        step: The step size. Must be positive.

    Returns:
        A tuple containing the quantized step number and the total number of steps in the range.
    """
    assert_true(stop > start, "stop must be strictly greater than start")
    assert_true(step > 0, "step must be positive")
    total_steps = _max(1, _ceil((stop - start) / step))
    result_steps = clamp(_round((value - start) / step), 0, total_steps - 1)
    return result_steps, total_steps


def make_comparable_float(*values: tuple[int, int]) -> float:
    """Convert a series of non-negative integer values into a float that compares the same way as the original series.

    This is useful for z-indexes, since Sonolus only supports a single float for z-index.

    The product of all maximum values must be less than 2^31.

    Usage:
        ```python
        make_comparable_float(
            quantize_to_step(time, -100, 100, 0.001),
            quantize_to_step(abs(lane), 0, 16, 0.01),
        )
        ```

    Args:
        *values: A series of tuples (value, max_value) where value falls in the range [0, max_value).

    Returns:
        A single float that can be used for comparison.
    """
    _check_max_total_steps(values)
    result = _ints_to_uint32(*values)
    return _uint32_to_comparable_float(result)


@meta_fn
def _check_max_total_steps(values: Iterable[tuple[int, int]]):
    result = 1
    for _, max_value in values:
        result *= max_value
    if is_static_true(result >= 2**31):
        raise ValueError(
            "The product of all maximum values must be less than 2^31. "
            "If using quantize_to_step, increase step sizes or reduce ranges."
        )


def product(values: Iterable[float]) -> float:
    """Calculate the product of an iterable of floats."""
    result = 1.0
    for value in values:
        result *= value
    return result


@meta_fn
def _ints_to_uint32(*values: tuple[int, int]) -> _UInt32:
    values = validate_value(values)
    if not isinstance(values, TupleImpl):
        raise TypeError("Expected a tuple of (value, max_value) pairs")
    values = values.value
    if len(values) == 0:
        return _UInt32.zero()
    collapsed_values = []
    running_value = validate_value(0)
    running_max_value = 1
    for value, max_value in values:
        value = validate_value(value)
        max_value = validate_value(max_value)
        if not max_value._is_py_():
            if running_max_value > 1:
                collapsed_values.append((running_value, running_max_value))
            collapsed_values.append((value, max_value))
            running_value = validate_value(0)
            running_max_value = 1
        elif running_max_value * max_value._as_py_() >= 2**24:
            if running_max_value > 1:
                collapsed_values.append((running_value, running_max_value))
            running_value = validate_value(value)
            running_max_value = max_value._as_py_()
        else:
            running_value = running_value * max_value._as_py_() + validate_value(value)
            running_max_value *= max_value._as_py_()
    if running_max_value > 1:
        collapsed_values.append((running_value, running_max_value))

    result = _UInt32.zero()
    for i, (value, steps) in enumerate(collapsed_values):
        if i == 0:
            result = _UInt32.of(value)
        else:
            result = result * _UInt32.of(steps) + _UInt32.of(value)
    return result


def _uint32_to_comparable_float(value: _UInt32) -> float:
    value = _UInt32._(value.hi, value.lo)
    if value < _HALF_MAX_TOTAL_STEPS_UINT32:
        sign = -1
        value = _HALF_MAX_TOTAL_STEPS_UINT32_MINUS_ONE - value
        hi = value.hi
        lo = value.lo
    else:
        sign = 1
        value = value - _HALF_MAX_TOTAL_STEPS_UINT32  # noqa: PLR6104
        hi = value.hi
        lo = value.lo
    exponent = hi // (2**7)
    mantissa = (hi % (2**7)) * (2**16) + lo
    return sign * (1 + mantissa / 2**23) * (2**exponent)



from typing import Any

type LocalizationText = dict[str, str]
type AnyText = str | LocalizationText


def as_localization_text(text: AnyText) -> LocalizationText:
    if isinstance(text, str):
        return {"en": text}
    return text


class Tag:
    """A tag for an engine or level.

    Args:
        title: The title of the tag.
        icon: The icon of the tag.
    """

    title: LocalizationText
    icon: str | None

    def __init__(self, title: AnyText, icon: str | None = None) -> None:
        self.title = as_localization_text(title)
        self.icon = icon

    def as_dict(self) -> dict[str, Any]:
        result = {"title": self.title}
        if self.icon is not None:
            result["icon"] = self.icon
        return result



from itertools import pairwise

from pydori.play.connector import HoldConnector, SimLine
from pydori.play.event import BpmChange, TimescaleChange
from pydori.play.note import (
    FlickNote,
    HoldAnchorNote,
    HoldEndNote,
    HoldHeadNote,
    HoldTickNote,
    Note,
    TapNote,
)
from pydori.play.stage import Stage
from sonolus.script.archetype import PlayArchetype
from sonolus.script.level import Level, LevelData


def demo_level():
    entities = [
        Stage(),
        BpmChange(
            beat=0,
            bpm=60,
        ),
        TapNote(
            beat=1,
            lane=0,
        ),
        TapNote(
            beat=2,
            lane=-2,
        ),
        TapNote(
            beat=3,
            lane=2,
        ),
        TapNote(
            beat=3,
            lane=0,
        ),
        TimescaleChange(
            beat=4.5,
            timescale=0.5,
        ),
        TapNote(
            beat=5,
            lane=-2,
        ),
        TapNote(
            beat=5,
            lane=0,
        ),
        TapNote(
            beat=5,
            lane=2,
        ),
        TimescaleChange(
            beat=6.5,
            timescale=0.2,
        ),
        BpmChange(
            beat=6,
            bpm=120,
        ),
        hold(
            HoldHeadNote(
                beat=7,
                lane=-3,
            ),
            HoldTickNote(
                beat=7.25,
                lane=-2,
            ),
            HoldTickNote(
                beat=7.5,
                lane=-1,
            ),
            HoldTickNote(
                beat=7.75,
                lane=0,
            ),
            HoldTickNote(
                beat=8,
                lane=1,
            ),
            HoldTickNote(
                beat=8.25,
                lane=2,
            ),
            HoldEndNote(
                beat=8.5,
                lane=3,
            ),
        ),
        TimescaleChange(
            beat=8.5,
            timescale=1.2,
        ),
        hold(
            HoldHeadNote(
                beat=9,
                lane=-3,
            ),
            HoldAnchorNote(
                beat=9.5,
                lane=3,
            ),
            FlickNote(
                beat=10,
                lane=-2,
            ),
        ),
    ]

    return Level(
        name="pydori_level",
        title="pydori Level",
        bgm=None,
        data=LevelData(
            bgm_offset=0,
            entities=[
                *entities,
                *create_sim_lines(entities),
            ],
        ),
    )


def hold(
    *notes: Note,
) -> list[PlayArchetype]:
    """Update the notes to reference each other and create connectors, then return notes and connectors in a list."""
    connectors = []
    for a, b in pairwise(sorted(notes, key=lambda n: n.beat)):
        connectors.append(HoldConnector(first_ref=a.ref(), second_ref=b.ref()))
        b.prev_ref = a.ref()
        a.next_ref = b.ref()
    return [
        *notes,
        *connectors,
    ]


def create_sim_lines(entities: list[PlayArchetype]) -> list[PlayArchetype]:
    """Create sim lines for the given entities."""
    if not entities:
        return []

    notes = [n for n in entities if isinstance(n, Note) and not isinstance(n, HoldTickNote | HoldAnchorNote)]

    notes.sort(key=lambda n: (n.beat, n.lane))
    sim_lines = []
    for a, b in pairwise(notes):
        if a.beat == b.beat:
            sim_lines.append(SimLine(first_ref=a.ref(), second_ref=b.ref()))

    return sim_lines



from sonolus.script.effect import StandardEffect, effects

# The same sfx played within this many seconds of the previous one will not be played,
# to prevent them from clashing and sounding unpleasant.
SFX_DISTANCE = 0.02


@effects
class Effects:
    # Empty tap.
    stage: StandardEffect.STAGE

    # Non-flick notes.
    perfect: StandardEffect.PERFECT
    great: StandardEffect.GREAT
    good: StandardEffect.GOOD

    # Flick notes.
    perfect_alt: StandardEffect.PERFECT_ALTERNATIVE
    great_alt: StandardEffect.GREAT_ALTERNATIVE
    good_alt: StandardEffect.GOOD_ALTERNATIVE

    # Active hold.
    hold: StandardEffect.HOLD



import pytest

from sonolus.script.debug import terminate
from sonolus.script.internal.context import ctx
from sonolus.script.internal.meta_fn import meta_fn
from tests.script.conftest import run_and_validate


def test_run_and_validate_error_if_returned_results_differ():
    @meta_fn
    def fn():
        if ctx():
            return 1
        else:
            return 2

    with pytest.raises(AssertionError):
        run_and_validate(fn)


def test_run_and_validate_error_if_only_py_raises():
    @meta_fn
    def fn():
        if ctx():
            return 1
        else:
            raise RuntimeError()

    with pytest.raises(AssertionError):
        run_and_validate(fn)


def test_run_and_validate_error_if_only_compiled_raises():
    @meta_fn
    def terminate_if_compiled():
        if ctx():
            terminate()

    def fn():
        terminate_if_compiled()
        return 1

    with pytest.raises(AssertionError):
        run_and_validate(fn)


def test_run_and_validate_raises_if_both_error():
    @meta_fn
    def fn():
        terminate()
        return 1

    with pytest.raises(RuntimeError, match="Terminated"):
        run_and_validate(fn)


def test_run_and_validate_success_if_both_return_same():
    @meta_fn
    def fn():
        return 1

    run_and_validate(fn)



from sonolus.backend.optimize.flow import (
    BasicBlock,
    traverse_cfg_reverse_postorder,
)
from sonolus.backend.optimize.passes import CompilerPass, OptimizerConfig


class DominanceFrontiers(CompilerPass):
    def destroys(self) -> set[CompilerPass] | None:
        return set()

    def run(self, entry: BasicBlock, config: OptimizerConfig) -> BasicBlock:
        blocks = list(traverse_cfg_reverse_postorder(entry))

        self.number_blocks(blocks)
        self.initialize_idoms(blocks, entry)
        self.compute_idoms(blocks)
        self.build_dominator_tree(blocks)
        self.compute_dominance_frontiers(blocks)

        return entry

    def number_blocks(self, blocks: list[BasicBlock]):
        """Assign a unique number to each block based on reverse post-order."""
        for idx, block in enumerate(blocks):
            block.num = idx

    def initialize_idoms(self, blocks: list[BasicBlock], entry_block: BasicBlock):
        """Initialize immediate dominators for each block."""
        for block in blocks:
            block.idom = None
        entry_block.idom = entry_block

    def compute_idoms(self, blocks: list[BasicBlock]):
        """Iteratively compute the immediate dominators of each block."""
        changed = True
        while changed:
            changed = False
            for b in blocks[1:]:  # Skip the entry block
                new_idom = None
                for edge in b.incoming:
                    p = edge.src
                    if p.idom is not None:
                        if new_idom is None:
                            new_idom = p
                        else:
                            new_idom = self.intersect(p, new_idom)
                if b.idom != new_idom:
                    b.idom = new_idom
                    changed = True

    def build_dominator_tree(self, blocks: list[BasicBlock]):
        """Construct the dominator tree using the immediate dominators."""
        for block in blocks:
            block.dom_children = []

        for block in blocks:
            if block.idom != block:
                block.idom.dom_children.append(block)

    def compute_dominance_frontiers(self, blocks: list[BasicBlock]):
        """Compute the dominance frontiers for all blocks."""
        for block in blocks:
            block.df = set()

        for b in blocks:
            if len(b.incoming) >= 2:
                for edge in b.incoming:
                    p = edge.src
                    runner = p
                    while runner != b.idom:
                        runner.df.add(b)
                        runner = runner.idom

    def intersect(self, b1: BasicBlock, b2: BasicBlock) -> BasicBlock:
        """Helper function to find the closest common dominator of two blocks."""
        while b1 != b2:
            while b1.num > b2.num:
                b1 = b1.idom
            while b2.num > b1.num:
                b2 = b2.idom
        return b1

    def __eq__(self, other):
        return isinstance(other, DominanceFrontiers)

    def __hash__(self):
        return hash(DominanceFrontiers)


def get_df(block: BasicBlock) -> set[BasicBlock]:
    return block.df


def get_dom_children(block: BasicBlock) -> list[BasicBlock]:
    return block.dom_children
